/***** spin: fun.c *****/

/* dSpin3.2 function definition and invocation routines                    */
/* Written by Radu Iosif (iosif@athena.polito.it)                          */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include <stdlib.h>
#include <string.h>

#include "spin.h"

#define STACK_QUANTA  1024


Defun *dfs = (Defun *)0;
int nrFun  = 0;
int nrPars = 0;

extern Symbol *context;
extern Symbol *Fname;
extern int lineno;
extern int nrRdy;
extern int Pid;

extern void put_seq(Sequence *, int, int);
extern int  doformal(FILE *, int, Sequence *, int, char *);
extern void fun_idx(Defun *);


Symbol *RetType = ZS;


void set_defun_proto(Lextok *nm, Lextok *prms, Lextok *rettype)
{
  Lextok *f, *p;
  Defun *df;

  for (df = dfs; df; df = df->nxt)
    if (strcmp(nm->sym->name, df->fun->name) == 0)
      break;

  if (!df) {
    df = (Defun *) emalloc(sizeof(struct Defun));
    df->fun = nm->sym;
    df->nxt = dfs;
    dfs     = df; 
  }
  else {
    if (df->body)
      fatal("duplicate definition of function '%s()'", df->fun->name);
  }

  for (f = prms; f; f = f->rgt)
    for (p = f->lft; p; p = p->rgt)
      {
	if (p->sym->type == STRUCT)
	  fatal("struct parameter for function '%s()'", df->fun->name);
      }

  df->prms = prms;
  df->rt   = rettype ? rettype->sym : ZS;
}


void set_defun_body(Lextok *nm, Lextok *bd)
{
  Defun *df;

  for (df = dfs; df; df = df->nxt)
    {
      if (strcmp(nm->sym->name, df->fun->name) == 0)
	break;
    }

  if (!df)
    fatal("cannot happen (set_defun_body)", (char *)0);

  df->body = bd->sq;
  df->body->cnxt = df->fun;

  fun_idx(df);
}


Defun *getdefun(Symbol *nm)
{
  Defun *df;

  for (df = dfs; df; df = df->nxt)
    if (strcmp(nm->name, df->fun->name) == 0)
      return df;

  return (Defun *)0;
}


int isdefun(char *s)
{
  Defun *df;

  for (df = dfs; df; df = df->nxt)
    if (strcmp(s, df->fun->name) == 0)
      return 1;

  return 0;
}


int get_func_address(Lextok *n)
{
  Defun *df;

  if (!(df = getdefun(n->sym)))
    {
      non_fatal("function '%s' not found", n->sym->name);
      return 0;
    }

  return (int)df->body;
}


void order_funs()
{
  Defun *df;
  Lextok *f, *g;
  int k, ord = nrRdy + 1; /* + np_ daemon */
  
  for (df = dfs; df; df = df->nxt)
    {
      for (f = df->prms, k = 0; f; f = f->rgt)
	for (g = f->lft; g; g = g->rgt, k ++);

      nrPars = max(nrPars, k);
      df->tn = ord ++;
      nrFun ++;
    }
}


void genfunpars()
{
  int i;

  for (i = 0; i < nrPars; i ++)
    {
      if (i == 0)
	fprintf(th, "\nint Par0");
      else
	fprintf(th, ", Par%d", i);
    }

  if (nrPars)
    fprintf(th, ";\n");
}


void putfun(Defun *df)
{
  Element *e;
  int ini;

  e   = df->body->frst;
  ini = huntele(e, e->status)->seqno;  
  Pid = df->tn;

  fprintf(th, "\nshort nstates%d = %d;\t/* %s */\n\n",
	  df->tn, df->body->maxel, df->fun->name);

  fprintf(th, "#define start%d\t\t%d\n", df->tn, ini);
  fprintf(th, "#define endstate%d\t%d\n\n",
	  df->tn, df->body->last->seqno);

  fprintf(tm, "\n\t\t/* FUN %s */\n", df->fun->name);
  fprintf(tb, "\n\t\t/* FUN %s */\n", df->fun->name);

  fprintf(tt, "\n\t\t/* defun %d: %s */\n", df->tn, df->fun->name);
  fprintf(tt, "\n  trans[%d] = (Trans **)", df->tn);
  fprintf(tt, " emalloc(%d* sizeof(Trans *));\n\n", df->body->maxel);

  RetType = df->rt;
  put_seq(df->body, 2, 0);
  RetType = ZS;

  dumpsrc(df->body->maxel, df->tn);
}


void genpushpars()
{
  Defun *df;
  int i, sz;

  fprintf(tc, "\ninline void pps(int h, int f");

  for (i = 0; i < nrPars; i ++)
    fprintf(tc, ", int p%d", i);

  fprintf(tc, ")\n{\n");
  fprintf(tc, "  switch (f) {\n");

  for (df = dfs; df; df = df->nxt)
    {
      fprintf(tc, "  case %d:\n", df->tn);

      sz = doformal(0, ASSIGN, df->body, 0, 0);

      __align(sz);

      if (sz)
	{
	  fprintf(tc, "\t\tpush(h, %d);\n", sz);
	  doformal(tc, INITVAR, df->body, 0, "p");
	}

      df->body->sz = sz;
      fprintf(tc, "  break;\n\n");
    }
  
  fprintf(tc, "  default:\n");
  fprintf(tc, "    Uerror(\"bad defun (pps)\");\n");
  fprintf(tc, "  }\n");
  fprintf(tc, "}\n\n\n");
}


RunStack *stack_init()
{
  RunStack *s = (RunStack *) emalloc(sizeof(struct RunStack));

  s->stack = (char *) emalloc(STACK_QUANTA);
  s->size  = STACK_QUANTA;
  s->used  = 0;
  
  return s;
}


char *stack_push(RunStack *s, int increment)
{
  char * sp;

  __align(increment);

  if (s->used + increment >= s->size) {
    s->size += STACK_QUANTA;
    s->stack = (char *) erealloc(s->stack, s->size);
  }

  sp = s->stack + s->used;
  s->used += increment;

  memset(sp, 0, increment);

  return sp;
}


void stack_pop(RunStack *s, int decrement)
{
  if (s->used < decrement)
    fatal("cannot happen -- stack_pop", (char *)0);

  __align(decrement);

  if (s->size - s->used + decrement > STACK_QUANTA) {
    s->size -= STACK_QUANTA;
    s->stack = (char *) erealloc(s->stack, s->size);
  }

  s->used -= decrement;
}


char *stack_peek(RunStack *s, int size)
{
  if (s->used < size)
    fatal("cannot happen -- stack_peek", (char *)0);

  return s->stack + s->used - size;
}


