/***** spin: pangen1.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include <string.h>

#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif
#include "pangen1.h"
#include "pangen3.h"

extern FILE	*tc, *th, *tt;
extern Label	*labtab;
extern UType    *Unames;
extern Ordered	*all_names;
extern ProcList	*rdy;
extern Defun    *dfs;
extern Queue	*qtab;
extern Symbol	*Fname;
extern int	lineno, verbose, Pid;
extern int	nrRdy, nrFun, nqs, mst, Mpars, claimnr, eventmapnr;
extern short	has_sorted, has_random, has_provided;
extern int      heap_init_size;

int	Npars=0, u_sync=0, u_async=0;

static Symbol	*LstSet=ZS;
static int	acceptors = 0, progressors = 0, nBits = 0;


static int Types[] = {
  UNSIGNED, INT, FTYPE, STRUCT,
  SHORT, MTYPE, CHAN, BYTE, BIT
};


static char *TNames[] = {
  "uint", "int", "uint", NULL,
  "short", "uchar", "uchar", "uchar", "uchar"
};


int             *plen_tab;
int             **pmap_tab;

int             *clen_tab;
int             **cmap_tab;

static void	dohidden(void);
static void     global_init(Symbol *, int);
static void     local_init(FILE *, Symbol *, int);
static void     gc_map(Symbol *, int, int);
static void     ch_map(Symbol *, int, int);
static void     gc_inc(FILE *, Symbol *, int);
static void     gc_inc_undo(FILE *, Symbol *, int);
static void     gc_dec(FILE *, Symbol *, int, int);
static void     gc_dec_undo(FILE *, Symbol *, int);
static void	do_init(FILE *, Symbol *);
static void	end_labs(Sequence *, int);
static void	put_ptype(int, int);
static void	tc_predef_np(void);
static void	put_pinit(ProcList *);
       void	walk_struct(int, char *, Symbol *, char *, char *, char *);

int	        doglobal(int);
int             doformal(FILE *, int, Sequence *, int, char *);
int             dolocal(FILE *, int, Sequence *, int);


static void 
reverse_names(ProcList *p)
{
  if (!p) 
    return;

  reverse_names(p->nxt);
  fprintf(th, "   \"%s\",\n", p->n->name);
}


static void
reverse_names2(Defun *d)
{
  if (!d)
    return;

  reverse_names2(d->nxt);
  fprintf(th, "   \"%s\",\n", d->fun->name);
}


void 
genheader(void)
{
  ProcList *p; int i;

  fprintf(th, "#define SYNC	%d\n", u_sync);
  fprintf(th, "#define ASYNC	%d\n\n", u_async);

  fprintf(th, "char *procname[] = {\n");
  reverse_names(rdy);
  fprintf(th, "   \":np_:\",\n");
  reverse_names2(dfs);
  fprintf(th, "};\n\n");

  put_ptype(mst, nrRdy + nrFun + 1);  
}


void genaddproc(void)
{	
  ProcList *p;
  int i;

  fprintf(tc, "int addproc(int n");

  for (i = 0; i < Npars; i++)
    fprintf(tc, ", int par%d", i);

  ntimes(tc, 0, 1, Addp0);
  ntimes(tc, 0, 1, Addp1);

  if (has_provided)
    {	
      fprintf(tt, "\nint provided(int II, int ot, ");
      fprintf(tt, "int tt, Trans *t)\n");
      fprintf(tt, "{\n\tswitch (ot) {\n");
    }
	
  tc_predef_np();
	
  for (p = rdy; p; p = p->nxt) 
    {
      Pid = p->tn;
      put_pinit(p);
    }

  Pid = 0;

  if (has_provided)
    {	
      fprintf(tt, "\tdefault: return 1; /* e.g., a claim */\n");
      fprintf(tt, "\t}\n\treturn 0;\n}\n");
    }

  ntimes(tc, i, i + 1, R6);
}


void genother(void)
{	
  ProcList *p;

  ntimes(tc, 0, 1, Code0);
  ntimes(tc, 0, nrRdy + nrFun + 1, R0); /* +1 for np_ */

  for (p = rdy; p; p = p->nxt)
    end_labs(p->s, p->tn);

  ntimes(tc, 0, nrRdy, R0a);
  ntimes(tc, nrRdy + 1, nrRdy + nrFun + 1, R0a);

  ntimes(tc, 0,     1, R0b);

  ntimes(th, acceptors,   acceptors + 1,   Code1);
  ntimes(th, progressors, progressors + 1, Code3);
  ntimes(th, nrRdy + nrFun + 1, nrRdy + nrFun + 2, R2); /* +1 for np_ */
  fprintf(tc, "  iniglobals();\n\n");
  ntimes(tc, 0, 1, Code2);
  ntimes(tc, 0, nrRdy, R4);
  fprintf(tc, "}\n\n");

  fprintf(tc, "void iniglobals(void)\n{\n");
  fprintf(tc, "  int h = -1, _t = Q_GLOBAL;\n");
  fprintf(tc, "#ifdef POSYMM\n");
  fprintf(tc, "  scalar_t *_s = (scalar_t *) 0;\n");
  fprintf(tc, "#endif /* POSYMM */\n\n");
  fprintf(tc, "  now._hsz = %d;\n", heap_init_size);
  doglobal(INITVAR);
  fprintf(tc, "}\n\n");

  doglobal(GC1_MAP);
  doglobal(CH_MAP);
}


static struct {
	char *s,	*t;		int n,	m,	p;
} ln[] = {
	{"end",  	"stopstate",	3,	0,	0},
	{"progress",	"progstate",	8,	0,	1},
	{"accept",	"accpstate",	6,	1,	0},
	{0,		0,		0,	0,	0},
};


static void
end_labs(Sequence *s, int i)
{	
  int oln = lineno;
  Symbol *ofn = Fname;
  Label *l;
  Symbol *sp;
  int j;

  for (sp = s->symtab; sp; sp = sp->next)
    if (sp->type == LABEL)
      for (l = labtab; l; l = l->nxt)
	for (j = 0; ln[j].n; j++) 
	  if (l->s == sp)
	    {
	      if (strncmp(l->s->name, ln[j].s, ln[j].n) == 0)
		{	
		  fprintf(tc, "  %s[%d][%d] = 1;\n", 
			  ln[j].t, i, l->e->seqno);
		  
		  acceptors += ln[j].m;
		  progressors += ln[j].p;
	    
		  if (l->e->status & D_ATOM)
		    {	
		      char foo[40];
		      
		      sprintf(foo, "%s-state label in d_step seq", ln[j].s);
		      lineno = l->e->n->ln;
		      Fname  = l->e->n->fn;
		      non_fatal("%s - ignored", foo);
		    }	
		}

	      /* visible states -- through remote refs: */	      
	      if (l->visible)
		fprintf(tc, "\tvisstate[%d][%d] = 1;\n", 
			i, l->e->seqno);
	    }

  lineno = oln;
  Fname  = ofn;
}


void
ntimes(FILE *fd, int n, int m, char *c[])
{
	int i, j;
	for (j = 0; c[j]; j++)
	for (i = n; i < m; i++)
	{	fprintf(fd, c[j], i, i, i, i, i, i);
		fprintf(fd, "\n");
	}
}

void
prehint(Symbol *s)
{	Lextok *n;

	printf("spin: warning, ");
	if (!s) return;

	n = (s->context != ZS)?s->context->ini:s->ini;
	if (n)
	printf("line %3d %s, ", n->ln, n->fn->name);
}


void
checktype(Symbol *sp, char *s)
{	
  char buf[32]; int i;

  if (!s
      || (sp->type != BYTE
	  &&  sp->type != SHORT
	  &&  sp->type != INT))
    return;

  if (sp->hidden&16)	/* formal parameter */
    {	
      ProcList *p; Lextok *f, *t;
      int posnr = 0;
	
      for (p = rdy; p; p = p->nxt)
	if (p->n->name
	    &&  strcmp(s, p->n->name) == 0)
	  break;
		
      if (p)
	for (f = p->p; f; f = f->rgt) /* list of types */
	  for (t = f->lft; t; t = t->rgt, posnr++)
	    if (t->sym
		&&  strcmp(t->sym->name, sp->name) == 0)
	      {	
		checkrun(sp, posnr);
		return;
	      }
    } 
  else 
    if (!(sp->hidden&4))
      {	
	if (!(verbose&32)) 
	  return;

	sputtype(buf, sp->type);
	i = strlen(buf);
	while (buf[--i] == ' ') buf[i] = '\0';
	prehint(sp);

	if (sp->context)
	  printf("proctype %s:", s);
	else
	  printf("global");

	printf(" '%s %s' could be declared 'bit %s'\n",
	       buf, sp->name, sp->name);
      } else 
	if (sp->type != BYTE && !(sp->hidden&8))
	  {	
	    if (!(verbose&32)) 
	      return;
	    
	    sputtype(buf, sp->type);
	    i = strlen(buf);
	    while (buf[--i] == ' ') buf[i] = '\0';
	    prehint(sp);
	    
	    if (sp->context)
	      printf("proctype %s:", s);
	    else
	      printf("global");
	    
	    printf(" '%s %s' could be declared 'byte %s'\n",
		   buf, sp->name, sp->name);
	  }
}


static void 
symbol_assign(Symbol *sp, unsigned int *offp, unsigned int *bitp)
{
  Lextok *fp, *tl;
  int ix;

  static int lst = 0;

  if (lst == BIT 
      && sp->type != BIT 
      && *bitp) 
    {
      (*offp) ++;
      (*bitp) = 0;
    }

  switch (sp->type)
    {
    case BIT:
      sp->offset  = *offp;
      sp->bitpos  = *bitp;
      sp->bitsize = 1;

      (*offp) += ((*bitp) + sp->nel) / 8;
      (*bitp)  = ((*bitp) + sp->nel) % 8;
      break;

    case BYTE:
    case CHAN:
    case MTYPE:
      sp->offset  = *offp;
      sp->bitpos  = 0;
      sp->bitsize = 8 * sizeof(char);

      (*offp) += sp->nel * sizeof(char);
      break;

    case SHORT:
      sp->offset  = *offp;
      sp->bitpos  = 0;
      sp->bitsize = 8 * sizeof(short);

      (*offp) += sp->nel * sizeof(short);
      break;

    case UNSIGNED:
    case INT:
    case FTYPE:
      __align(*offp);

      sp->offset  = *offp;
      sp->bitpos  = 0;
      sp->bitsize = 8 * sizeof(int);

      (*offp) += sp->nel * sizeof(int);

      break;

    case STRUCT:
      {
	int ix  = 0;
	int j;
	
	__align(*offp);

	for (fp = sp->Slst; fp; fp = fp->rgt)
	  for (tl = fp->lft; tl; tl = tl->rgt)
	    {
	      symbol_assign(tl->sym, &ix, bitp);
	    }

	if (*bitp) {
	  (*bitp) = 0;
	  ix ++;
	}

	sp->offset  = *offp;
	sp->bitpos  = 0;
	sp->bitsize = ix * 8; 

	(*offp) += sp->nel * ix;

	break;
      }

    default:
      if (sp->type & POINTER) 
	{
	  __align(*offp);

	  sp->offset  = *offp;
	  sp->bitpos  = 0;
	  sp->bitsize = 8 * sizeof(int);

	  (*offp) += sp->nel * sizeof(int);
	}
      else
	fatal("undeclared variable %s\n", sp->name);
    }

  lst = sp->type;
}


static void
unames_assign(void)
{
  UType *tmp;
  int off, bit;

  for (tmp = Unames; tmp; tmp = tmp->nxt)
    {
      off = bit = 0;
      symbol_assign(tmp->nm, &off, &bit);
    }
}


static void
fields_init(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;
	    
	    switch (s->type)
	      {
	      case STRUCT:
		fields_init(s, newb + s->offset);
		break;

	      case BIT:
		if (s->nel == 1)
		  {
		    fprintf(tc, "\tbset(gg(h_offset(k)+base+%d, char), %d, ",
			    newb + s->offset, s->bitpos);
		    do_init(tc, s);
		    fprintf(tc, "); /* %s */\n", s->name);
		  }
		else
		  for (i = 0; i < s->nel; i ++)
		    {
		      fprintf(tc, "\tbset(gg(h_offset(k)+base+%d, char), %d, ",
			      ((newb + s->offset) * 8 + s->bitpos + i) / 8,
			      ((newb + s->offset) * 8 + s->bitpos + i) % 8);
		      do_init(tc, s);
		      fprintf(tc, "); /* %s */\n", s->name);
		    }

		break;

	      default:
		for (i = 0; i < s->nel; i ++) {
		  fprintf(tc, "\tgg(h_offset(k)+base+%d, %s) = ",
			  newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		  do_init(tc, s);
		  fprintf(tc, "; /* %s */\n", s->name);
		}
	      }
	  }  
    }
}


static void
fields_gc0(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;
	    switch (s->type)
	      {
	      case STRUCT:
		fields_gc0(s, newb + s->offset);
		break;
		
	      default:
		if (s->type & POINTER)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\t__gc_dec(gg(h_offset(k)+base+%d, %s), 0);",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, " /* %s[%d].%s[%d] */\n", 
			    sp->name, ix, s->name, i);
		  }
		break;
	      }
	  }
    }
}


static void
fields_gc1(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;

	    switch (s->type)
	      {
	      case STRUCT:
		fields_gc1(s, newb + s->offset);
		break;
		
	      default:
		if (s->type & POINTER)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\tgc_mark(gg(h_offset(k)+base+%d, %s));",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, " /* %s[%d].%s[%d] */\n", 
			    sp->name, ix, s->name, i);
		  }
		break;
	      }
	  }
    }
}


static void
fields_ptr_reloc(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;

	    switch (s->type)
	      {
	      case STRUCT:
		fields_ptr_reloc(s, newb + s->offset);
		break;
		
	      default:
		if (s->type & POINTER)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\t/* %s[%d].%s[%d] */\n\t", 
			    sp->name, ix, s->name, i);
		    fprintf(tc, "i = h_index(gg(h_offset(k)+base+%d,%s));\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, "#ifndef NOSYMM\n\t");
		    fprintf(tc, "if (i >= threshold)\n");
		    fprintf(tc, "\t  gg(h_offset(k)+base+%d,%s)",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, " = mp(i+inc, 0,  HPTR);\n\t");		    
		    fprintf(tc, "else\n");
		    fprintf(tc, "#endif /* NOSYMM */\n\t");
		    fprintf(tc, "if (del >= 0 && i == del)\n");
		    fprintf(tc, "\t  gg(h_offset(k)+base+%d,%s) = DPTR;\n\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		  }
		
		break;
	      }
	  }
    }
}


static void
fields_chan_reloc(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;

	    switch (s->type)
	      {
	      case STRUCT:
		fields_chan_reloc(s, newb + s->offset);
		break;
		
	      default:
		if (s->type == CHAN)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\t/* %s[%d].%s[%d] */\n", 
			    sp->name, ix, s->name, i);
		    fprintf(tc, "#ifndef NOSYMM\n\t");
		    fprintf(tc,"if (gg(h_offset(k)+base+%d,%s)>=threshold)\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, "\t  gg(h_offset(k)+base+%d,%s) += inc;\n\t",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, "else\n");
		    fprintf(tc, "#endif /* NOSYMM */\n\t");
		    fprintf(tc, "if (del && gg(h_offset(k)+base+%d,%s)",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, " == del)\n");
		    fprintf(tc, "\t  gg(h_offset(k)+base+%d,%s)=0;\n\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		  }
		
		break;
	      }
	  }
    }
}


static void
fields_ptr_perm(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;

	    switch (s->type)
	      {
	      case STRUCT:
		fields_ptr_perm(s, newb + s->offset);
		break;
		
	      default:
		if (s->type & POINTER)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\t/* %s[%d].%s[%d] */\n\t", 
			    sp->name, ix, s->name, i);
		    fprintf(tc,"i=h_index(gg(h_offset(k)+base+%d,%s));\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, "\tif (0 <= i && i < l)\n");
		    fprintf(tc, "\t  gg(h_offset(k)+base+%d,%s)",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, " = mp(p[i], 0, HPTR);\n\n");
		  }
		
		break;
	      }
	  }
    }
}


static void
fields_chan_perm(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb, i;
  Symbol *s;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    s = tl->sym;

	    switch (s->type)
	      {
	      case STRUCT:
		fields_chan_perm(s, newb + s->offset);
		break;
		
	      default:
		if (s->type == CHAN)
		  for (i = 0; i < s->nel; i ++) {
		    fprintf(tc, "\t/* %s[%d].%s[%d] */\n\t", 
			    sp->name, ix, s->name, i);
		    fprintf(tc,"i=gg(h_offset(k)+base+%d,%s)-1;\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		    fprintf(tc, "\tif (0 <= i && i < l)\n");
		    fprintf(tc, "\tgg(h_offset(k)+base+%d,%s)=p[i]+1;\n\n",
			    newb + s->offset + i * (s->bitsize / 8), typ2c(s));
		  }
		
		break;
	      }
	  }
    }
}


static void
formal_init(FILE *fd, Symbol *sp, char *par)
{
  switch (sp->type)
    {
    case STRUCT:
      fprintf(fd, "\t\tstruct_copy((char *)lptr(%d), (char *)%s%d, %d);\n",
	      sp->offset, par, sp->parno, sp->bitsize / 8);
      break;
      
    case BIT:
      fprintf(fd, "\t\tbset(ll(%d, char), %d, %s%d);\n",
	      sp->offset, sp->bitpos, par, sp->parno);
      break;

    default:
      fprintf(fd, "\t\tll(%d, %s) = %s%d;\n",
	      sp->offset, typ2c(sp), par, sp->parno);
    }
}


static void
local_struct_init(FILE *fd, Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      local_struct_init(fd, tl->sym, newb + tl->sym->offset);
	    else
	      local_init(fd, tl->sym, newb);
	  }
    }
}


static void
local_init(FILE *fd, Symbol *sp, int base)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      local_struct_init(fd, sp, base + sp->offset);
      break;

    case BIT:
      if (sp->nel == 1)
	{
	  fprintf(fd, "\t\tbset(ll(%d, char), %d, ",
		  base + sp->offset, sp->bitpos);
	  do_init(fd, sp);
	  fprintf(fd, ");");
	}
      else
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(fd, "\t\tbset(ll(%d, char), %d, ",
		    ((base + sp->offset) * 8 + sp->bitpos + i) / 8,
		    ((base + sp->offset) * 8 + sp->bitpos + i) % 8);
	    do_init(fd, sp);
	    fprintf(fd, ");");
	  }

      break;

    default:
      for (i = 0; i < sp->nel; i ++)
	{
	  fprintf(fd, "\t\tll(%d, %s) = ",
		  base + sp->offset + i * (sp->bitsize / 8), typ2c(sp));
	  do_init(fd, sp);
	  fprintf(fd, "; /* %s */", sp->name);	  
	}
    }

  fprintf(fd, "\n");
}


static void
struct_gc_dec(FILE *fd, Symbol *sp, int base, int save)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_gc_dec(fd, tl->sym, newb + tl->sym->offset, save);
	    else
	      gc_dec(fd, tl->sym, newb, save);
	  }
    }
}


static void
gc_dec(FILE *fd, Symbol *sp, int base, int save)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_gc_dec(fd, sp, base + sp->offset, save);
      break;

    default:
      if (sp->type & POINTER)
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(fd, "\t\tgc_dec(ll(%d, %s), %d); /* %s */\n",
		    base + sp->offset + i * (sp->bitsize / 8), 
		    typ2c(sp), save, sp->name);
	  }
    }
}


static void
struct_gc_dec_undo(FILE *fd, Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_gc_dec_undo(fd, tl->sym, newb + tl->sym->offset);
	    else
	      gc_dec_undo(fd, tl->sym, newb);
	  }
    }
}


static void
gc_dec_undo(FILE *fd, Symbol *sp, int base)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_gc_dec_undo(fd, sp, base + sp->offset);
      break;

    default:
      if (sp->type & POINTER)
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(fd, "\t\tgc_dec_back(ll(%d, %s)); /* %s */\n",
		    base + sp->offset + i * (sp->bitsize / 8), 
		    typ2c(sp), sp->name);
	  }
    }
}


static void
struct_gc_inc(FILE *fd, Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_gc_inc(fd, tl->sym, newb + tl->sym->offset);
	    else
	      gc_inc(fd, tl->sym, newb);
	  }
    }
}


static void
gc_inc(FILE *fd, Symbol *sp, int base)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_gc_inc(fd, sp, base + sp->offset);
      break;

    default:
      if (sp->type & POINTER)
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(fd, "\t\tgc_inc(ll(%d, %s)); /* %s */\n",
		    base + sp->offset + i * (sp->bitsize / 8), 
		    typ2c(sp), sp->name);
	  }
    }
}


static void
struct_gc_inc_undo(FILE *fd, Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_gc_inc_undo(fd, tl->sym, newb + tl->sym->offset);
	    else
	      gc_inc_undo(fd, tl->sym, newb);
	  }
    }
}


static void
gc_inc_undo(FILE *fd, Symbol *sp, int base)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_gc_inc_undo(fd, sp, base + sp->offset);
      break;

    default:
      if (sp->type & POINTER)
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(fd, "\t\tgc_inc_back(ll(%d, %s)); /* %s */\n",
		    base + sp->offset + i * (sp->bitsize / 8), 
		    typ2c(sp), sp->name);
	  }
    }
}


static void
struct_gc_map(Symbol *sp, int base, int glob)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_gc_map(tl->sym, newb + tl->sym->offset, glob);
	    else
	      gc_map(tl->sym, newb, glob);
	  }
    }
}


static void
gc_map(Symbol *sp, int base, int glob)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_gc_map(sp, base + sp->offset, glob);
      break;

    default:
      if (sp->type & POINTER)
	for (i = 0; i < sp->nel; i ++)
	  {
	    int k = glob ? 0 : Pid + 1;

	    plen_tab[k] ++;

	    pmap_tab[k] = 
	      (int *) erealloc(pmap_tab[k], plen_tab[k] * sizeof(int));

	    pmap_tab[k][plen_tab[k] - 1] = 
	      base + sp->offset + i * (sp->bitsize / 8);
	  }
    }
}


static void
struct_ch_map(Symbol *sp, int base, int glob)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++)
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT)
	      struct_ch_map(tl->sym, newb + tl->sym->offset, glob);
	    else
	      ch_map(tl->sym, newb, glob);
	  }
    }  
}


static void
ch_map(Symbol *sp, int base, int glob)
{
  int i;

  switch (sp->type)
    {
    case STRUCT:
      struct_ch_map(sp, base + sp->offset, glob);
      break;

    default:
      if (sp->type == CHAN)
	for (i = 0; i < sp->nel; i ++)
	  {
	    int k = glob ? 0 : Pid + 1;

	    clen_tab[k] ++;

	    cmap_tab[k] = 
	      (int *) erealloc(cmap_tab[k], clen_tab[k] * sizeof(int));

	    cmap_tab[k][clen_tab[k] - 1] = 
	      base + sp->offset + i * (sp->bitsize / 8);
	  }
    }
}


int
doformal(FILE *fd, int dowhat, Sequence *s, int base, char *par)
{
  Symbol *sp;
  int k = 0;
  int j;

  unsigned int value  = base;
  unsigned int bitpos = 0;

  for (j = 0; j < 9; j ++)
    for (sp = s->symtab; sp; sp = sp->next)
      {
	if (sp->type == LABEL)
	  continue;

	if ((sp->hidden & 16)
	    && ((j == 0 && sp->type & POINTER) ||
		(sp->type == Types[j])))
	  switch (dowhat)
	    {
	    case ASSIGN:
	      symbol_assign(sp, &value, &bitpos);
	      break;

	    case INITVAR:
	      checktype(sp, (char *)0);
	      formal_init(fd, sp, par);
	      break;

	    case GC0_SAVE:
	      gc_dec(fd, sp, 0, 1);
	      break;

	    case GC0_NOSV:
	      gc_dec(fd, sp, 0, 0);
	      break;

	    case GC0_DEC_UNDO:
	      gc_dec_undo(fd, sp, 0);
	      break;

	    case GC0_INCR:
	      gc_inc(fd, sp, 0);
	      break;

	    case GC0_INC_UNDO:
	      gc_inc_undo(fd, sp, 0);
	      break;

	    case GC1_MAP:
	      gc_map(sp, 0, 0);
	      break;

	    case CH_MAP:
	      ch_map(sp, 0, 0);
	      break;
	    }
      }

  if (bitpos)
    value ++;

  return value - base;
}


int 
dolocal(FILE *fd, int dowhat, Sequence *s, int base)
{	
  Symbol *sp;
  int k = 0;
  int j;

  unsigned int value  = base;
  unsigned int bitpos = 0;

  for (j = 0; j < 9; j ++)
    for (sp = s->symtab; sp; sp = sp->next)
      {
	if (sp->type == LABEL)
	  continue;

	if (!(sp->hidden & 1)
	    && !(sp->hidden & 16) 
	    && ((j == 0 && sp->type & POINTER) ||
		(sp->type == Types[j])))
	  {
	    switch (dowhat)
	      {
	      case ASSIGN:
		symbol_assign(sp, &value, &bitpos);
		break;
		
	      case INITVAR:
		checktype(sp, (char *)0);
		local_init(fd, sp, 0);
		break;

	      case GC0_INCR:
		gc_inc(fd, sp, 0);
		break;

	      case GC0_INC_UNDO:
		gc_inc_undo(fd, sp, 0);
		break;

	      case GC0_SAVE:
		gc_dec(fd, sp, 0, 1);
		break;

	      case GC0_NOSV:
		gc_dec(fd, sp, 0, 0);
		break;

	      case GC0_DEC_UNDO:
		gc_dec_undo(fd, sp, 0);
		break;

	      case GC1_MAP:
		gc_map(sp, 0, 0);
		break;

	      case CH_MAP:
		ch_map(sp, 0, 0);
		break;
	      }
	  }
      }

  if (bitpos)
    value ++;

  return value - base;
}


static void
dostruct()
{
  UType *u;
  int off, bit;

  for (u = Unames; u; u = u->nxt)
    {
      off = bit = 0;
      symbol_assign(u->nm, &off, &bit);
    }
}


int 
global_assign()
{
  dostruct();
  dohidden();
  return doglobal(ASSIGN);
}


static void
global_struct_init(Symbol *sp, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  for (ix = 0; ix < sp->nel; ix ++) 
    {
      newb = base + ix * (sp->bitsize / 8);

      for (fp = sp->Slst; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {
	    if (tl->sym->type == STRUCT) 
	      global_struct_init(tl->sym, newb + tl->sym->offset);
	    else
	      global_init(tl->sym, newb);
	  }
    }
}


static void 
global_init(Symbol *sp, int base)
{	
  int i;

  switch (sp->type)
    {
    case STRUCT:
      /* struct may contain a chan */     
      global_struct_init(sp, base + sp->offset);
      break;

    case BIT:
      
      if (sp->nel == 1)
	{
	  fprintf(tc, "  bset(gg(%d, char), %d, ", 
		  base + sp->offset, sp->bitpos);	      
	  do_init(tc, sp);
	  fprintf(tc, ");\n");
	}
      else
	for (i = 0; i < sp->nel; i ++)
	  {
	    fprintf(tc, "  bset(gg(%d, char), %d, ",
		    ((base + sp->offset) * 8 + sp->bitpos + i) / 8, 
		    ((base + sp->offset) * 8 + sp->bitpos + i) % 8);
	    do_init(tc, sp);
	    fprintf(tc, ");\n");
	  }

      break;

    default:

      for (i = 0; i < sp->nel; i++)
	{	
	  fprintf(tc, "  gg(%d, %s) = ", 
		  base + sp->offset + i * (sp->bitsize / 8), typ2c(sp));
	  do_init(tc, sp);
	  fprintf(tc, ";\n");
	}

      break;
    }
}


int 
doglobal(int dowhat)
{	
  Ordered *walk;
  Symbol *sp;
  int j;

  unsigned int value = 0;
  unsigned int bitpos = 0;

  for (j = 0; j < 9; j ++)
    for (walk = all_names; walk; walk = walk->next)
      {	
	sp = walk->entry;

	if (!sp->context
	    && !sp->owner
	    && !(sp->hidden & 1)
	    && !(sp->hidden & 128)
	    && ((j == 0 && sp->type & POINTER) || 
		(sp->type == Types[j])))
	  {	
	    if (Types[j] != MTYPE || !ismtype(sp->name))
	      switch (dowhat) 
		{
		case ASSIGN:
		  symbol_assign(sp, &value, &bitpos);
		  break;
		  
		case INITVAR:
		  checktype(sp, (char *) 0);
		  global_init(sp, 0);
		  value ++; 
		  break;

		case GC1_MAP:
		  gc_map(sp, 0, 1);
		  break;
		  
		case CH_MAP:
		  ch_map(sp, 0, 1);
		  break;
		}
	  }
      }
  
  if (bitpos)
    value ++;

  return value;
}


static void
dohidden(void)
{	
  Ordered *walk;
  Symbol *sp;
  int j;
  int ok = 0;

  for (j = 0; j < 9; j++)
    for (walk = all_names; walk; walk = walk->next)
      {	
	sp = walk->entry;

	if (!(sp->hidden & 128)
	    && (sp->hidden & 1)
	    &&  sp->type == Types[j])
	  {	
	    if (sp->context || sp->owner)
	      fatal("cannot hide non-globals (%s)", sp->name);

	    if (sp->type == CHAN)
	      fatal("cannot hide channels (%s)", sp->name);

	    if (sp->type == STRUCT)
	      fatal("dSpin cannot hide structures (%s)", sp->name);

	    if (! ok++)
	      fprintf(th, "\n\n/*** Hidden variables ***/\n");
	    
	    fprintf(th, "%s %s", typ2c(sp), sp->name);
	    
	    if (sp->nel > 1)
	      fprintf(th, "[%d]", sp->nel);

	    fprintf(th, ";\n");
	  }
      }
}


static void
do_init(FILE *fd, Symbol *sp)
{	
  int i; extern Queue *ltab[];

  if (sp->ini
      &&  sp->type == CHAN
      && ((i = qmake(sp)) > 0))
    {	
      if (sp->ini->ntyp == CHAN) {
	fprintf(fd, "\n#ifdef POSYMM");
	fprintf(fd, "\naddqueue(_t, h, %d, %d, _s)", i, ltab[i-1]->nslots==0);
	fprintf(fd, "\n#else");
	fprintf(fd, "\naddqueue(_t, h, %d, %d)", i, ltab[i-1]->nslots==0);
	fprintf(fd, "\n#endif /* POSYMM */\n");
      }
      else
	fprintf(fd, "%d", i);
    } 
  else
    if (sp->ini
	&& sp->ini->ntyp == NAME) 
      putname(fd, "", sp->ini, 0, "", !(node_type2(sp->ini) & POINTER));
    else
      putstmnt(fd, sp->ini, 0);
}


static int blog(int n)	/* for small log2 without rounding problems */
{	
  int m = 1, r = 2;

  while (r < n) 
    { 
      m ++; 
      r *= 2; 
    }
	
  return 1 + m;
}


static char *type_tab[] = { "uchar", "ushort", "uint" };
static int  size_tab[]  = { sizeof(char), sizeof(short), sizeof(int) };


static int bmap(int n)
{
  if (n < (1 << (8 * sizeof(char))))
    return 0;

  if (n < (1 << (8 * sizeof(short))))
    return 1;
  
  return 2;
}


static void put_ptype(int m0, int m1)
{
  fprintf(th, "/*** Process type ***/\n\n");
  fprintf(th, "#ifndef STACKSZ\n");
  fprintf(th, "#define STACKSZ 256\n");
  fprintf(th, "#endif /* STACKSZ */\n\n");

  fprintf(th, "typedef struct P {\n");

#ifdef __ALIGN
  fprintf(th, "  uchar _pid;  /* 0..255 */\n");    
  fprintf(th, "  %s _t;  /* proctype */\n",   type_tab[bmap(m1)]);
  fprintf(th, "  %s _p;  /* state    */\n\n", type_tab[bmap(m0)]);
#else
  fprintf(th, "  unsigned _pid : 8;  /* 0..255 */\n");
  fprintf(th, "  unsigned _t   : %d; /* proctype */\n",   blog(m1));
  fprintf(th, "  unsigned _p   : %d; /* state    */\n\n", blog(m0));
  fprintf(th, "  uchar pad[0];\n\n");
#endif /* __ALIGN */

  fprintf(th, "#define ot_t %s\n",   type_tab[bmap(m1)]);
  fprintf(th, "#define tt_t %s\n\n", type_tab[bmap(m0)]);

  fprintf(th, "#if STACKSZ < 65536\n");
  fprintf(th, "#define bp_t short\n");

#ifdef __ALIGN
  fprintf(th, "  uchar pad[%d];\n",
	  (WORD - (2 + size_tab[bmap(m0)] + 2 * size_tab[bmap(m1)] +
		   2 * sizeof(short)) % WORD) % WORD);
#endif /* __ALIGN */

  fprintf(th, "#else\n");
  fprintf(th, "#define bp_t long\n");

#ifdef __ALIGN
  fprintf(th, "  uchar pad[%d];\n",
	  (WORD - (2 + size_tab[bmap(m0)] + 2 * size_tab[bmap(m1)] +
		   2 * sizeof(long)) % WORD) % WORD);
#endif /* __ALIGN */

  fprintf(th, "#endif\n\n");

  fprintf(th, "  %s _type;\n\n", type_tab[bmap(m1)]);

  fprintf(th, "  unsigned bp_t _sp; /* stack pointer */\n");
  fprintf(th, "  unsigned bp_t _bp; /* base pointer */\n");
  fprintf(th, "  unsigned char _rf; /* &1 = rv pointer, &2 = rv chan */\n");
  fprintf(th, "  int _rv;           /* return value */\n");

  fprintf(th, "  uchar stk[STACKSZ]; /* run-time stack */\n");
  fprintf(th, "} P;\n\n");
}


static void tc_predef_np(void)
{	
  int i = nrRdy;	/* 1+ highest proctype nr */

  fprintf(th, "#define _NP_\t%d\n", i);
  fprintf(th, "uchar reached%d[3];  \t/* np_ */\n", i);
  fprintf(th, "short nstates%d = 3; \t/* np_ */\n", i);
  fprintf(th, "#define endstate%d\t2\t/* np_ */\n\n", i);
  fprintf(th, "#define start%d\t0   \t/* np_ */\n", i);

  fprintf(tc, "  case %d:	/* np_ */\n", i);
  fprintf(tc, "    this->_p = 0;\n", i);
  fprintf(tc, "    reached%d[0] = 1;\n", i);
  fprintf(tc, "    accpstate[%d][1] = 1;\n", i);
  fprintf(tc, "    break;\n");
}


static void put_pinit(ProcList *P)
{	
  Lextok	*fp, *fpt, *t;
  Element	*e = P->s->frst;
  Symbol	*s = P->n;
  Lextok        *p = P->p;

  int	i = P->tn;
  int	ini, j, k;
  
  ini = huntele(e, e->status)->seqno;
  fprintf(th, "#define start%d\t%d\n", i, ini);

  if (i == claimnr)
    fprintf(th, "#define start_claim	%d\n", ini);

  if (i == eventmapnr)
    fprintf(th, "#define start_event	%d\n", ini);

  fprintf(tc, "  case %d:	/* %s */\n", i, s->name);
  fprintf(tc, "    this->_p = %d;\n", ini);
  fprintf(tc, "    reached%d[%d] = 1;\n", i, ini);

  if (has_provided)
    {	
      fprintf(tt, "\tcase %d: /* %s */\n\t\t", i, s->name);
      if (P->prov)
	{	
	  fprintf(tt, "if (");
	  putstmnt(tt, P->prov, 0);
	  fprintf(tt, ")\n\t\t\t");
	}

      fprintf(tt, "return 1;\n");
      if (P->prov)
	fprintf(tt, "\t\tbreak;\n");
    }

  fprintf(tc, "    this->_sp = 0;\n");
  fprintf(tc, "    this->_bp = 0;\n");

  if (P->s->sz)
    {
      fprintf(tc, "    push(h, %ld);\n", P->s->sz);

      fprintf(tc, "\n /* formals */\n");
      doformal(tc, INITVAR, P->s, 0, "par");
      doformal(0, GC1_MAP, P->s, 0, 0);
      doformal(0, CH_MAP, P->s, 0, 0);

      fprintf(tc, "\n /* locals */\n");
      enter_seq(P->s);
      dolocal(tc, INITVAR, P->s, 0);
      dolocal(0, GC1_MAP, P->s, 0);
      dolocal(0, CH_MAP, P->s, 0);
      
      exit_seq();
    }

  enter_seq(P->s);
  dumpclaims(tc, i, s->name);
  exit_seq();

  fprintf(tc, "    break;\n");
}


Element *
huntstart(Element *f)
{	Element *e = f;

	if (e->n)
	{	if (e->n->ntyp == '.' && e->nxt)
			e = e->nxt;
		else if (e->n->ntyp == ATOMIC
		     ||  e->n->ntyp == D_STEP
		     ||  e->n->ntyp == NON_ATOMIC)
			e->n->sl->this->last->nxt = e->nxt;
		else if (e->n->ntyp == UNLESS)
			return e->sub->this->frst;
	}
	return e;
}

Element *
huntele(Element *f, int o)
{	
  Element *g, *e = f;
  int cnt; /* a precaution against loops */

  if (e)
    for (cnt=0; cnt < 10 && e->n; cnt++)
      {	
	switch (e->n->ntyp) {
	case GOTO:
	  g = get_lab(e->n,1);
	  cross_dsteps(e->n, g->n);
	  break;

	case '.':
	case BREAK:
	  if (!e->nxt)
	    return e;

	  g = e->nxt;
	  break;

	case UNLESS:
	  g = huntele(e->sub->this->frst, o);
	  break;
		
	case D_STEP:
	case ATOMIC:
	case NON_ATOMIC:
	  e->n->sl->this->last->nxt = e->nxt;

	default:	/* fall through */
	  return e;
	}
	
	if ((o & ATOM) && !(g->status & ATOM))
	  return e;

	e = g;
      }
  
  return e;
}


char * typ2c(Symbol *sp)
{	
  char buf[1024];

  switch (sp->type) 
    {
    case UNSIGNED:
      sprintf(buf, "uint");
      break;

    case BIT:
    case MTYPE:
    case CHAN:	/* good for up to 255 channels */
      sprintf(buf, "uchar");
      break;

    case BYTE:
      sprintf(buf, "char");
      break;

    case SHORT:
      sprintf(buf, "short");
      break;

    case INT:
      sprintf(buf, "int");
      break;

    case PREDEF:
      return;

    case FTYPE:
      sprintf(buf, "uint");
      break;

    case STRUCT:
      fatal("cannot happen: typ2c(%s)", sp->name); 

    default:
      if (sp->type & POINTER)
	sprintf(buf, "uint");
      else
	fatal("variable %s undeclared", sp->name);
    }

  return (char *)strdup(buf);
}


char * dep2c(Symbol *sp)
{
  Symbol *vp = (Symbol *) emalloc(sizeof(struct Symbol));

  memcpy(vp, sp, sizeof(struct Symbol));
  vp->type &= ~POINTER;

  if (vp->type == STRUCT)
    return (char *)strdup("uint");

  return typ2c(vp);
}


static void ncases(FILE *fd, int p, int n, int m, char *c[])
{	
  int i, j;
  
  for (j = 0; c[j]; j++)
    for (i = n; i < m; i++) {	
      fprintf(fd, c[j], i, p, i);
      fprintf(fd, "\n");
    }
}


void genaddheap(void)
{
  int j, type = 0;
  UType *un;

  ntimes(tc, 0, 1, Addh0);

  fprintf(tc, "    default: if (t & %d) {\n", POINTER);
  fprintf(tc, "               sz = sizeof(uint);\n");
  fprintf(tc, "               type = %d;\n", type ++);
  fprintf(tc, "               break;\n");
  fprintf(tc, "             }\n\n");
  fprintf(tc, "             uerror(\"bad type (addheap)\");\n");
  fprintf(tc, "             break;\n\n");

  for (j = 0; j < 9; j ++)
    {
      if (Types[j] != STRUCT)
	fprintf(tc, "    case %d:\t\tsz = sizeof(%s);\ttype = %d;\tbreak;\n",
		Types[j], TNames[j], type ++);
    }

  unames_assign();
  fprintf(tc, "  /* user defined types */\n");

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "    case %d:\tsz = %d;\ttype = %d;\tbreak;\t/* %s */\n",
	      un->nm->Nid, un->nm->bitsize / 8, type, un->nm->name); 
      un->rt_type = type ++;
    }

  ntimes(tc, 0, 1, Addh1);
  fprintf(tc, "  /* initializations */\n");

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->nm->Nid, un->nm->name);
      fields_init(un->nm, 0);
      fprintf(tc, "      break;\n\n");
    }

  ntimes(tc, 0, 1, Addh2);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_gc0(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }

  ntimes(tc, 0, 1, Addh3);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_gc1(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }

  ntimes(tc, 0, 1, Addh4);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_ptr_reloc(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }
  
  ntimes(tc, 0, 1, Addh5);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_chan_reloc(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }

  ntimes(tc, 0, 1, Addh6);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_ptr_perm(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }

  ntimes(tc, 0, 1, Addh7);

  for (un = Unames; un; un = un->nxt)
    {
      fprintf(tc, "      case %d:\t/* %s */\n", un->rt_type, un->nm->name);
      fields_chan_perm(un->nm, 0);
      fprintf(tc, "      base += %d; break;\n\n", un->nm->bitsize / 8);
    }  

  ntimes(tc, 0, 1, Addh8);
}


void genaddqueue(void)
{	
  char buf0[256];
  int j;
  Queue *q;

  ntimes(tc, 0, 1, Addq0);

  fprintf(th, "\nshort q_flds[%d];\n", nqs + 1);
  fprintf(th, "short q_max[%d];\n\n", nqs + 1);

  for (q = qtab; q; q = q->nxt)
    {	
      j = q->qid;

      fprintf(tc, "  case %d:\n", j);
      fprintf(tc, "    j = sizeof(Q%d);\n", j);
      fprintf(tc, "    q_flds[%d] = %d;\n", j, q->nflds);
      fprintf(tc, "    q_max[%d] = %d;\n", j, max(1,q->nslots));

#ifdef __ALIGN
      if (q->align)
	fprintf(tc, "    align = 1;\n");
#endif /* __ALIGN */

      fprintf(tc, "    break;\n\n");

      ntimes(th, j, j+1, R9);

      for (j = 0; j < q->nflds; j++)
	{	
	  if (q->fld_width[j] & POINTER)
	    {
	      fprintf(th, "    uint fld%d;\n", j);
	    }
	  else switch (q->fld_width[j])
	    {
	    case BIT:
	      if (q->nflds != 1)
		{	
		  fprintf(th, "    unsigned fld%d : 1;\n", j);
		  break;
		} /* else fall through: smaller struct */

	    case MTYPE:
	    case CHAN:
	    case BYTE:
	      fprintf(th, "    uchar fld%d;\n", j);
	      break;

	    case FTYPE:
	      fprintf(th, "    uint fld%d;\n", j);
	      break;

	    case SHORT:
	      fprintf(th, "    short fld%d;\n", j);
	      break;

	    case INT:
	      fprintf(th, "    int fld%d;\n", j);
	      break;

	    default:
	      printf("%d: ", q->fld_width[j]);
	      fatal("bad channel spec", (char *)0);
	    }
	}

      fprintf(th, "  } contents[%d];\n", max(1, q->nslots));
      fprintf(th, "} Q%d;\n\n", q->qid);
    }

  ntimes(th, 0, 1, R10);
  ntimes(tc, 0, 1, Addq1);

  if (has_random)
    {	
      fprintf(tc, "int Q_has(int into");

      for (j = 0; j < Mpars; j++)
	fprintf(tc, ", int want%d, int fld%d", j, j);

      fprintf(tc, ")\n");
      fprintf(tc, "{");
      fprintf(tc, "  int i;\n\n");
      fprintf(tc, "  if (!into--)\n");
      fprintf(tc, "    uerror(\"ref to unknown chan (recv-poll)\");\n\n");
      fprintf(tc, "  if (into < 0 || q_free(into))\n");
      fprintf(tc, "    Uerror(\"qrecv bad queue#\");\n\n");
      fprintf(tc, "	for (i = 0; i < qptr(into)->Qlen;");
      fprintf(tc, " i++)\n");
      fprintf(tc, "	{\n");

      for (j = 0; j < Mpars; j++)
	{	
	  fprintf(tc, "		if (want%d && ", j);
	  fprintf(tc, "qrecv(into+1, i, %d, 0) != fld%d)\n",
		  j, j);
	  fprintf(tc, "			continue;\n");
	}

      fprintf(tc, "		return i+1;\n");
      fprintf(tc, "	}\n");
      fprintf(tc, "	return 0;\n");
      fprintf(tc, "}\n");
    }

  fprintf(tc, "void qsend(int into, int sorted");

  for (j = 0; j < Mpars; j++)
    fprintf(tc, ", int fld%d", j);

  fprintf(tc, ")\n");
  ntimes(tc, 0, 1, Addq11);
  
  for (q = qtab; q; q = q->nxt)
    {	
      sprintf(buf0, "((Q%d *)z)->", q->qid);
      fprintf(tc, "  case %d : %s\n", q->qid, (q->nslots) ? "" : "/* =rv= */");

      if (q->nslots == 0)	/* reset handshake point */
	fprintf(tc, "    (trpt+2)->o_m = 0;\n");

      if (has_sorted)
	{	
	  fprintf(tc, "    if (!sorted)\n");
	  fprintf(tc, "      goto append%d;\n\n", q->qid);
	  fprintf(tc, "    for (j = 0; j < %sQlen; j++) {\n", buf0);
	  sprintf(buf0, "((Q%d *)z)->contents[j].fld", q->qid);

	  for (j = 0; j < q->nflds; j++)
	    {	
	      fprintf(tc, "      if (fld%d > %s%d) continue;\n", j, buf0, j);
	      fprintf(tc, "      if (fld%d < %s%d)\n", j, buf0, j);
	      fprintf(tc, "        goto found%d;\n\n", q->qid);
	    }
		
	  fprintf(tc, "    }\n");
	  fprintf(tc, "found%d:\n", q->qid);
	  sprintf(buf0, "((Q%d *)z)->", q->qid);

	  fprintf(tc, "    for (k = %sQlen - 1; k >= j; k--) {\n", buf0);
	  for (j = 0; j < q->nflds; j++)
	    {	
	      fprintf(tc, "      %scontents[k+1].fld%d = ", buf0, j);
	      fprintf(tc, "      %scontents[k].fld%d;\n", buf0, j);
	    }

	  fprintf(tc, "    }\n");
	  fprintf(tc, "append%d:\t/* insert in slot j */\n", q->qid);
	}

      fprintf(tc, "    (trpt+1)->oval = j;\n");
      fprintf(tc, "    %sQlen = %sQlen + 1;\n", buf0, buf0);
      sprintf(buf0, "((Q%d *)z)->contents[j].fld", q->qid);

      for (j = 0; j < q->nflds; j++)
	fprintf(tc, "    %s%d = fld%d;\n", buf0, j, j);

      fprintf(tc, "    break;\n\n");
    }
	
  ntimes(tc, 0, 1, Addq2);

  for (q = qtab; q; q = q->nxt)
    fprintf(tc, "  case %d: return %d;\n", q->qid, (!q->nslots));

  ntimes(tc, 0, 1, Addq3);

  for (q = qtab; q; q = q->nxt)
    fprintf(tc, "  case %d: return (qptr(from)->Qlen == %d);\n",
	    q->qid, max(1, q->nslots));

  ntimes(tc, 0, 1, Addq4);

  for (q = qtab; q; q = q->nxt)
    {	
      sprintf(buf0, "((Q%d *)z)->", q->qid);
      fprintf(tc, "  case %d:%s\n", q->qid, (q->nslots)? "":"/* =rv= */");

      if (q->nflds == 1) {
	fprintf(tc, "    if (fld == 0)\n");
	fprintf(tc, "      r = %scontents[slot].fld0;\n\n", buf0);
      }
      else {	
	fprintf(tc, "    switch (fld) {\n");
	ncases(tc, q->qid, 0, q->nflds, R12);
	fprintf(tc, "    default: Uerror(\"too many fields in recv\");\n");
	fprintf(tc, "    }\n\n");
      }
		
      fprintf(tc, "    if (done) {\n");
      fprintf(tc, "      j = %sQlen;\n",  buf0);
      fprintf(tc, "      %sQlen = -- j;\n\n", buf0);
      fprintf(tc, "      for (k = slot; k < j; k ++) {\n");
      sprintf(buf0, "((Q%d *)z)->contents", q->qid);

      for (j = 0; j < q->nflds; j ++)
	fprintf(tc, "       %s[k].fld%d = %s[k+1].fld%d;\n", buf0, j, buf0, j);

      fprintf(tc, "      }\n\n");

      for (j = 0; j < q->nflds; j++)
	fprintf(tc, "    %s[j].fld%d = 0;\n", buf0, j);

      fprintf(tc, "\n    if (fld+1 != %d)\n", q->nflds);
      fprintf(tc, "      uerror(\"missing pars in receive\");\n");
      /* incompletely received msgs cannot be unrecv'ed */
      fprintf(tc, "    }\n");
      fprintf(tc, "    break;\n\n");
    }
  
  ntimes(tc, 0, 1, Addq5);

  fprintf(th, "void qsend(int, int");
  for (j = 0; j < Mpars; j++)
    fprintf(th, ", int");
  fprintf(th, ");\n\n");

  fprintf(th, "#define Addproc(x)\taddproc(x");
  for (j = 0; j < Npars; j++)
    fprintf(th, ", 0");
  fprintf(th, ")\n\n");
}
