/***** spin: spin.y *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


%{
#include "spin.h"
#include <stdarg.h>

#define YYDEBUG	   1

#define Stop	    nn(ZN, STOP, ZN, ZN)


extern  Symbol	*context, *owner;
extern  int	u_sync, u_async, dumptab, analyze, lineno;
extern	short	has_sorted, has_random, has_enabled, has_pcvalue, has_np;
extern SeqList  *cur_s;

extern	void	validref(Lextok *, Lextok *);
extern  int     samename(Symbol *, Symbol *);
extern  Symbol *addsym(char *);
extern  char   *type_name(int);

extern  void    check_unames();
extern  void    order_funs();


int	Mpars = 0;	/* max nr of message parameters  */
int     ParNo;          /* formal parameter number */
int	runsafe = 1;	/* 1 if all run stmnts are in init */
int     run_ok = 0;     /* flag = 1 only in init{} */
int	Expand_Ok = 0, realread = 1, IArgs = 0, NamesNotAdded = 0, Parm = 0;

char	*claimproc = (char *) 0;
char	*eventmap  = (char *) 0;

static  Lextok  *FuncRetType = ZN;
static	int	 Embedded = 0, inEventMap = 0, has_ini = 0;
static  int      parse_errors = 0;
static  int      return_ok = 0;
static  int      unless_block = 0;
%}

%token	ASSERT PRINT
%token	RUN LEN ENABLED EVAL PC_VAL
%token	TYPEDEF INLINE LABEL OF
%token	GOTO BREAK ELSE SEMI
%token	IF FI DO OD SEP
%token	ATOMIC NON_ATOMIC D_STEP UNLESS 
%token  TIMEOUT NONPROGRESS
%token	ACTIVE PROCTYPE D_PROCTYPE HIDDEN SHOW FUNCTION
%token	PRIORITY PROVIDED
%token	FULL EMPTY NFULL NEMPTY
%token	CONST TYPE XU			/* val */
%token	NAME UNAME PNAME INAME FNAME	/* sym */
%token	STRING CLAIM TRACE INIT		/* sym */
%token  NEW DELETE __NULL RETURN
%token  FUNCALL RETFUN                  /* not used in production rules */

%right	ASGN
%left	SND O_SND RCV R_RCV /* SND doubles as boolean negation */
%left	OR
%left	AND
%left	'|'
%left	'^'
%left	'&'
%left	EQ NE
%left	GT LT GE LE
%left	LSHIFT RSHIFT
%left	'+' '-'
%left	'*' '/' '%'
%left	INCR DECR
%right	'~' UMIN NEG LVAL
%left	DOT

%expect 1
%%


/** PROMELA2 Grammar Rules **/

program	: units		{ extern char yytext[];
			  yytext[0] = '\0';

			  if (parse_errors)
			    return;

			  /* CHANGE: check out typedefs */
			  check_unames();

			  /* CHANGE: set functions ordinal number */
			  order_funs();

			  sched();
			}
	;

units	: unit
	| units unit
	;

unit	: proc		/* proctype { }        */
	| init		/* init { }            */
	| claim		/* never claim         */
	| events	/* event assertions    */
	| one_decl	/* variables, chans    */
	| utype		/* user defined types  */
	| ns		/* named sequence      */
        | defun         /* function definition */ 
	| SEMI		/* optional separator  */
	| error
	;

proc	: inst		/* optional instantiator */
	  proctype NAME	{
			  setptype($3, PROCTYPE, ZN);
			  setpname($3);

			  /* linenr and file */

			  Expand_Ok ++; /* expand struct names in decl */
			  has_ini = 0;

			  open_seq($3->sym->name, 1, 1, 0);

			  Parm ++;
			  ParNo = 0;
			}
	  '(' decl ')'	{ Expand_Ok --;
	                  Parm --;
			  if (has_ini)
			    fatal("initializer in parameter list", (char *) 0);
			}
	  Opt_priority
	  Opt_enabler
	  procbody      { ProcList *rl;
			  rl = ready($3->sym, $6, $11->sq, $2->val, $10);

			  if ($1 != ZN && $1->val > 0)
			    {
			      extern RunList *run;
			      int j;
			      
			      for (j = 0; j < $1->val; j++) 
				{				
				  runnable(rl, $9 ? $9->val : 1, 1);
				  if (!analyze) 
				    {
				      rt_enter(run, rl->s, LOCAL_TAG);
				      initlocals(run->rt_seq->this);
				    }
				}

			      announce(":root:");

			      if (dumptab) 
				$3->sym->ini = $1;
			    }
			  
			  context = ZS;
			}
	;

proctype: PROCTYPE	{ $$ = nn(ZN,CONST,ZN,ZN); $$->val = 0; }
	| D_PROCTYPE	{ $$ = nn(ZN,CONST,ZN,ZN); $$->val = 1; }
	;

inst	: /* empty */	       { $$ = ZN; }
	| ACTIVE	       { $$ = nn(ZN,CONST,ZN,ZN); $$->val = 1; }
	| ACTIVE '[' CONST ']' { $$ = nn(ZN,CONST,ZN,ZN); $$->val = $3->val; }
	;

init	: INIT           { run_ok ++;}
	  Opt_priority
	  body		 { ProcList *rl;
	                   extern RunList * run;

			   rl = ready($1->sym, ZN, $4->sq, 0, ZN);
			   runnable(rl, $3 ? $3->val : 1, 1);

			   announce(":root:");

			   if (!analyze)
			     {
			       rt_enter(run, rl->s, LOCAL_TAG);
			       initlocals(run->rt_seq->this);
			     }

			   context = ZS;
			   run_ok --;
	                  }
	;

claim	: CLAIM		{ /* context = $1->sym; */
			  if (claimproc)
			    non_fatal("claim %s redefined", claimproc);

			  claimproc = $1->sym->name;
			}

	  body		{ (void) ready($1->sym, ZN, $3->sq, 0, ZN);
        		  context = ZS;
        		}
	;

events : TRACE		{ /* context = $1->sym; */
			  if (eventmap)
			    non_fatal("trace %s redefined", eventmap);

			  eventmap = $1->sym->name;
			  inEventMap++;
			}
	  body		{ (void) ready($1->sym, ZN, $3->sq, 0, ZN);
        		  context = ZS;
			  inEventMap--;
			}
	;

utype	: TYPEDEF NAME  { if (context)
                            fatal("typedef %s must be global", $2->sym->name);

                          owner = $2->sym;
			  setuname(ZN);
			  owner = ZS;
                        }

        | TYPEDEF NAME	{ if (context)
	                    fatal("typedef %s must be global", $2->sym->name);

                          owner = $2->sym;
			  setuname(ZN);
                    	}

       '{' decl_lst '}' { setuname($5); 
                          owner = ZS; 
                        }

/* CHANGE: typedef already declared in a typedef prototype */
        | TYPEDEF UNAME { if (context)
	                    fatal("typedef %s must be global", $2->sym->name);

	                   owner = $2->sym;
	                }

       '{' decl_lst '}' { setuname($5); 
                          owner = ZS; 
                        }
        ;

nm	: NAME		 { $$ = $1; }
	| INAME		 { $$ = $1;

	                   if (IArgs)
			     fatal("invalid use of '%s'", $1->sym->name);
	                 }
	;

ns	: INLINE nm '('	 { NamesNotAdded++; }
	  args ')'	 { prep_inline($2->sym, $5); NamesNotAdded--; }
	;

/* CHANGE: function declarations */
fn      : NAME           
        | FNAME          
        ;

defun   : FUNCTION fn     { set_defun_proto($2, ZN, ZN); }
        | FUNCTION fn     { Expand_Ok ++;

                            if (context)
			      fatal("defun '%s()' must be global", 
				    $2->sym->name);

			    open_seq($2->sym->name, 1, 1, 1);
			    Parm ++;
			    ParNo = 0;
                           }
      '(' decl ')' rettype { set_defun_proto($2, $5, $7); 
	                     Expand_Ok --;
			     Parm --;
			     FuncRetType = $7;
			     return_ok ++;
	                   }
       funbody             { set_defun_body($2, $9); 
	                     FuncRetType = (Lextok *)0;
			     return_ok --;
	                   }
        ;

funbody : '{'                   { rem_local(); }
              sequence OS '}'   { Element *f, *l;
                                  Sequence *s;
				  
				  s = close_seq(1, 1);
 
				  f = frst_el(s);
				  l = last_el(s);

				  f->tag |= FB;
				  l->tag |= FE;

				  f->fsq = s;
				  l->fsq = s;

				  $$->sq = s;
				  unrem_local();
                                }
        ;

procbody : '{'                  { rem_local(); }
               sequence OS '}'  { add_seq(Stop);
                                  $$->sq = close_seq(1, 1);
				  unrem_local();
                                }
         ;

body	: '{'			{ rem_local();
		                  open_seq(0, 1, 1, 0); }
          sequence OS '}'	{ add_seq(Stop);
          			  $$->sq = close_seq(1, 1);
				  unrem_local();
	                        }
	;

block   : '{'                   { rem_local();
	                          if (unless_block)
                                    open_seq(0, 0, 0, 0);
                                  else
				    open_seq(0, 0, 1, 1);
                                }

          sequence OS '}'       { Element *f, *l;
	                          Sequence *s;

				  if (unless_block)
				    s = close_seq(0, 0);
				  else
				    s = close_seq(0, 1);

				  f = frst_el(s);
				  l = last_el(s);

				  f->tag |= BB;
				  f->sbl = seqlist(s, f->sbl);

				  if (!(l->tag & JMP)
				      && l->n->ntyp != RETURN
				      && l->n->ntyp != RETFUN)
				    {
				      l->tag |= BE;
				      l->sel = seqlist(s, l->sel);
				    }
				  
				  $$->sq = s;
				  unrem_local();
	                        }
        ;

atomic_block : '{'              { open_seq(0, 0, 0, 0); }
          sequence OS '}'       { $$->sq = check_seq(close_seq(0, 0), 7); }
        ;

sequence: step			{ if ($1) add_seq($1); }
	| sequence MS step	{ if ($3) add_seq($3); }
	;

step    : one_decl		{ $$ = ZN; }
	| XU vref_lst		{ setxus($2, $1->val); $$ = ZN; }
	| stmnt			{ $$ = $1; }
        | stmnt UNLESS          { unless_block ++; }
          stmnt	                { $$ = do_unless($1, $4); 
	                          unless_block --;
	                        }
	;

vis	: /* empty */		{ $$ = ZN; }
	| HIDDEN		{ $$ = $1; }
	| SHOW			{ $$ = $1; }
	;

asgn    : /* empty */
	| ASGN
	;

one_decl: vis TYPE var_list	{ setptype($3, $2->val, $1); $$ = $3; }

	| vis UNAME var_list	{ setutype($3, $2->sym, $1);
				  $$ = expand($3, Expand_Ok);
				}

	| vis TYPE asgn '{' nlst '}' {
				  if ($2->val != MTYPE)
				    fatal("malformed declaration", 0);
				  
				  setmtype($5);

				  if ($1)
				    non_fatal("cannot %s mtype (ignored)",
					      $1->sym->name);
				  
				  if (context != ZS)
				    fatal("mtype declaration must be global", 
					  0);
				}
	;

decl_lst: one_decl       	{ $$ = nn(ZN, ',', $1, ZN); }
	| one_decl SEMI
	  decl_lst		{ $$ = nn(ZN, ',', $1, $3); }
	;

decl    : /* empty */		{ $$ = ZN; }
	| decl_lst      	{ $$ = $1; }
	;

vref_lst: varref		{ $$ = nn($1, XU, $1, ZN); }
	| varref ',' vref_lst	{ $$ = nn($1, XU, $1, $3); }
	;

var_list: ivar           	{ $$ = nn($1, NAME, ZN, ZN); }
	| ivar ',' var_list	{ $$ = nn($1, NAME, ZN, $3); }
	;

ivar    : vardcl           	{ $$ = $1; }
	| vardcl ASGN expr   	{ $1->sym->ini = $3; $$ = $1;
				  trackvar($1,$3); has_ini = 1;
				}
	| vardcl ASGN ch_init	{ $1->sym->ini = $3;
				  $$ = $1; has_ini = 1;
				}
	;

ch_init : '[' CONST ']' OF
	  '{' typ_list '}'	{ if ($2->val) u_async++;
				  else u_sync++;
        			  {	
				    int i = cnt_mpars($6);
				    Mpars = max(Mpars, i);
				  }
				  
        			  $$ = nn(ZN, CHAN, ZN, $6);
				  $$->val = $2->val;
        			}
	;

vardcl: var_decl        
        ;

var_decl: NAME  		{ if ($1->sym->context != context)
                                    $$->sym = addsym($1->sym->name);
                                  else
				    $$->sym = $1->sym;

                                  $$->sym->nel = 1; 
				  if (Parm)
				    {
				      $$->sym->hidden |= 16;
				      $$->sym->parno = ParNo ++;
				    }
                                }

	| NAME ':' CONST	{ if ($1->sym->context != context)
	                            $$->sym = addsym($1->sym->name);
                                  else
				    $$->sym = $1->sym;

	                          $$->sym->nbits = $3->val;
				  $$->sym->nel = 1; 
				  if (Parm)
				    {
				      $$->sym->hidden |= 16;
				      $$->sym->parno = ParNo ++;
				    }
				}

	| NAME '[' CONST ']'	{ if ($1->sym->context != context)
	                            $$->sym = addsym($1->sym->name);
                                  else
				    $$->sym = $1->sym;

                                  $$->sym->nel = $3->val;
	                          if (Parm) 
				    {
				      $$->sym->hidden |= 16;
				      $$->sym->parno = ParNo ++;
				    }
	                        }

        | '&' var_decl          { $$->sym = $2->sym;
	                          $$->sym->type |= POINTER; 
                                }
	;

varref	: cmpnd			{ $$ = mk_explicit($1, Expand_Ok, NAME); }
        | FNAME                 { $$ = nn($1, FNAME, ZN, ZN); }
	;

pfld	: NAME			{ $$ = nn($1, NAME, ZN, ZN); }
	| NAME			{ owner = ZS; }
	  '[' expr ']'		{ $$ = nn($1, NAME, $4, ZN); }
	;

cmpnd	: pfld			{ Embedded++;
				  if ($1->sym->type & STRUCT)
					owner = $1->sym->Snm;
				}
	  sfld			{ $$ = $1; $$->rgt = $3;

				  if (!Embedded && !$1->sym->type)
				   non_fatal("undeclared variable: %s",
						$1->sym->name);
				  Embedded--;
				  /* if ($3) validref($1, $3->lft); */
				  owner = ZS;
				}
	;

sfld	: /* empty */		{ $$ = ZN; }
	| '.' cmpnd %prec DOT	{ $$ = nn(ZN, '.', $2, ZN); }
	;

stmnt	: Special		{ $$ = $1; }
	| Stmnt			{ $$ = $1;
				  if (inEventMap)
				   fatal("not an event", (char *)0);
				}
	;

Special : varref RCV		{ Expand_Ok++; }
	  rargs			{ Expand_Ok--;
				  $$ = nn($1,  'r', $1, $4);
				  trackchanuse($4, ZN, 'R');
				}

	| varref SND		{ Expand_Ok++; }
	  margs			{ Expand_Ok--;
				  $$ = nn($1, 's', $1, $4);
				  $$->val=0; trackchanuse($4, ZN, 'S');
				}

	| IF options FI 	{ $$ = nn($1, IF, ZN, ZN);
        			  $$->sl = $2->sl;
        			}

	| DO    		{ pushbreak(); }
          options OD    	{ $$ = nn($1, DO, ZN, ZN);
        			  $$->sl = $3->sl;
        			}

	| BREAK  		{ $$ = nn(ZN, GOTO, ZN, ZN);
				  $$->sym = break_dest();
				}

	| GOTO NAME		{ $$ = nn($2, GOTO, ZN, ZN);

				  if ($2->sym->type != 0
				      &&  $2->sym->type != LABEL) 
				    {
				      non_fatal("bad label-name %s",
						$2->sym->name);
				    }

				  $2->sym->type = LABEL;
				}

        | RETURN expr           { SeqList *l;

	                          if (!return_ok)
	                            fatal("'return' must occur within a defun",
					  (char *) 0);

	                          if (!FuncRetType)
				    fatal("return value from %s",
					  "void function");

	                          $$ = nn(FuncRetType, RETURN, $2, ZN);
				  $$->ln = lineno;

				  for (l = cur_s; l; l = l->nxt)
				    $$->sl = seqlist(l->this, $$->sl);
	                        }

        | RETURN                { SeqList *l;
	  
	                          if (!return_ok)
	                            fatal("'return' must occur within a defun",
					  (char *) 0);

	                          if (FuncRetType)
				    fatal("return without value from %s",
					  "non void function");

	                          $$ = nn(FuncRetType, RETURN, ZN, ZN); 
				  $$->ln = lineno;

				  for (l = cur_s; l; l = l->nxt)
				    $$->sl = seqlist(l->this, $$->sl);
	                        }

	| NAME ':' stmnt	{ $$ = nn($1, ':', $3, ZN);

				  if ($1->sym->type != 0
				      &&  $1->sym->type != LABEL) {
				    non_fatal("bad label-name %s",
					      $1->sym->name);
				  }

				  $1->sym->type = LABEL;
				}
	;

Stmnt	: varref ASGN expr	{ $$ = nn(ZN, ASGN, $1, $3);
				  trackvar($1, $3);
				  nochan_manip($1, $3, 0);
				}

        | varref ASGN funcall   { $$ = nn(ZN, FUNCALL, $1, $3); }
        | funcall               { $$ = nn(ZN, FUNCALL, ZN, $1); }
        | RETURN funcall        { $$ = nn(ZN, RETFUN, ZN, $2); }

	| varref INCR		{ $$ = nn(ZN,CONST, ZN, ZN); $$->val = 1;
				  $$ = nn(ZN,  '+', $1, $$);
				  $$ = nn($1, ASGN, $1, $$);
				  trackvar($1, $1);
				  if ($1->sym->type == CHAN)
				   fatal("arithmetic on chan", (char *)0);
				}
	| varref DECR		{ $$ = nn(ZN,CONST, ZN, ZN); $$->val = 1;
				  $$ = nn(ZN,  '-', $1, $$);
				  $$ = nn($1, ASGN, $1, $$);
				  trackvar($1, $1);
				  if ($1->sym->type == CHAN)
				   fatal("arithmetic on chan id's", (char *)0);
				}
	| PRINT	'(' STRING	{ realread = 0; }
	  prargs ')'		{ $$ = nn($3, PRINT, $5, ZN); realread = 1; }
	| ASSERT full_expr    	{ $$ = nn(ZN, ASSERT, $2, ZN); }
	| varref R_RCV		{ Expand_Ok++; }
	  rargs			{ Expand_Ok--;
				  $$ = nn($1,  'r', $1, $4);
				  $$->val = has_random = 1;
				  trackchanuse($4, ZN, 'R');
				}
	| varref RCV		{ Expand_Ok++; }
	  LT rargs GT		{ Expand_Ok--;
				  $$ = nn($1, 'r', $1, $5);
				  $$->val = 2;	/* fifo poll */
				  trackchanuse($5, ZN, 'R');
				}
	| varref R_RCV		{ Expand_Ok++; }
	  LT rargs GT		{ Expand_Ok--;	/* rrcv poll */
				  $$ = nn($1, 'r', $1, $5);
				  $$->val = 3; has_random = 1;
				  trackchanuse($5, ZN, 'R');
				}
	| varref O_SND		{ Expand_Ok++; }
	  margs			{ Expand_Ok--;
				  $$ = nn($1, 's', $1, $4);
				  $$->val = has_sorted = 1;
				  trackchanuse($4, ZN, 'S');
				}
	| full_expr		{ $$ = nn(ZN, 'c', $1, ZN); }
	| ELSE  		{ $$ = nn(ZN,ELSE,ZN,ZN);
				}

	| ATOMIC  atomic_block	{ 
	                          $$ = nn($1, ATOMIC, ZN, ZN);
        			  $$->sl = seqlist(check_seq($2->sq, 3), 0);
        			  make_atomic($$->sl->this, 0);
        			}

	| D_STEP 		{ rem_Seq(); }
          atomic_block 	        { $$ = nn($1, D_STEP, ZN, ZN);
        			  $$->sl = seqlist(check_seq($3->sq, 4), 0);
        			  make_atomic($$->sl->this, D_ATOM);
				  unrem_Seq();
        			}

	| block                 { $$ = nn(ZN, NON_ATOMIC, ZN, ZN);
        			  $$->sl = seqlist(check_seq($1->sq, 5), 0);
        			}

	| INAME			{ IArgs++; }
	  '(' args ')'		{ pickup_inline($1->sym, $4); IArgs--; }
	  Stmnt			{ $$ = $7; }

        | varref ASGN new       { $$ = nn(ZN, NEW, $1, $3); }
        | DELETE '(' varref ')' { $$ = nn(ZN, DELETE, $3, ZN); }
	;

options : option		{ $$->sl = seqlist($1->sq, 0); }
	| option options	{ $$->sl = seqlist($1->sq, $2->sl); }
	;

option  : SEP   		{ open_seq(0, 0, 0, 0); }
          sequence OS		{ $$->sq = close_seq(6, 0); }
	;

OS	: /* empty */
	| SEMI			{ /* redundant semi at end of sequence */ }
	;

MS	: SEMI			{ /* at least one semi-colon */ }
	| MS SEMI		{ /* but more are okay too   */ }
	;

aname	: NAME			{ $$ = $1; }
	| PNAME			{ $$ = $1; }
	;

funcall : varref '('             { Expand_Ok ++;
	                          if (!context)
				    fatal("function called outside proctype",
					  (char *)0);
                             	}
          args ')'              { Expand_Ok --;
	                          $$ = nn($1, $1->ntyp, $4, $1);
	                        } 
        ;

expr    : '(' expr ')'		{ $$ = $2; }
	| expr '+' expr		{ $$ = nn(ZN, '+', $1, $3); }
	| expr '-' expr		{ $$ = nn(ZN, '-', $1, $3); }
	| expr '*' expr		{ $$ = nn(ZN, '*', $1, $3); }
	| expr '/' expr		{ $$ = nn(ZN, '/', $1, $3); }
	| expr '%' expr		{ $$ = nn(ZN, '%', $1, $3); }
	| expr '&' expr		{ $$ = nn(ZN, '&', $1, $3); }
	| expr '^' expr		{ $$ = nn(ZN, '^', $1, $3); }
	| expr '|' expr		{ $$ = nn(ZN, '|', $1, $3); }
	| expr GT expr		{ $$ = nn(ZN,  GT, $1, $3); }
	| expr LT expr		{ $$ = nn(ZN,  LT, $1, $3); }
	| expr GE expr		{ $$ = nn(ZN,  GE, $1, $3); }
	| expr LE expr		{ $$ = nn(ZN,  LE, $1, $3); }
	| expr EQ expr		{ $$ = nn(ZN,  EQ, $1, $3); }
	| expr NE expr		{ $$ = nn(ZN,  NE, $1, $3); }
	| expr AND expr		{ $$ = nn(ZN, AND, $1, $3); }
	| expr OR  expr		{ $$ = nn(ZN,  OR, $1, $3); }
	| expr LSHIFT expr	{ $$ = nn(ZN, LSHIFT,$1, $3); }
	| expr RSHIFT expr	{ $$ = nn(ZN, RSHIFT,$1, $3); }
	| '~' expr		{ $$ = nn(ZN, '~', $2, ZN); }
	| '-' expr %prec UMIN	{ $$ = nn(ZN, UMIN, $2, ZN); }
	| SND expr %prec NEG	{ $$ = nn(ZN, '!', $2, ZN); }

	| '(' expr SEMI expr ':' expr ')' {
				  $$ = nn(ZN,  OR, $4, $6);
				  $$ = nn(ZN, '?', $2, $$);
				}

	| RUN aname		{ Expand_Ok++;

				  if (!context)
				    fatal("used 'run' outside proctype",
					  (char *) 0);

				  if (!run_ok)
				    runsafe = 0;
				}
	  '(' args ')'
	  Opt_priority		{ Expand_Ok--;
				  $$ = nn($2, RUN, $5, ZN);
				  $$->val = ($7) ? $7->val : 1;
				  trackchanuse($5, $2, 'A'); trackrun($$);
				}
	| LEN '(' varref ')'	{ $$ = nn($3, LEN, $3, ZN); }
	| ENABLED '(' expr ')'	{ $$ = nn(ZN, ENABLED, $3, ZN);
				  has_enabled++;
				}
	| varref RCV		{ Expand_Ok++; }
	  '[' rargs ']'		{ Expand_Ok--;
				  $$ = nn($1, 'R', $1, $5);
				}
	| varref R_RCV		{ Expand_Ok++; }
	  '[' rargs ']'		{ Expand_Ok--;
				  $$ = nn($1, 'R', $1, $5);
				  $$->val = has_random = 1;
				}
	| varref		{ $$ = $1; trapwonly($1, "varref"); }

        | '&'                   { Expand_Ok ++; }
          varref %prec LVAL     { $$ = nn($3, LVAL, $3, ZN); 
	                          Expand_Ok --;
	                        } 

        | __NULL                { $$ = nn($1, __NULL, ZN, ZN); }
	| CONST 		{ $$ = nn(ZN,CONST,ZN,ZN);
				  $$->ismtyp = $1->ismtyp;
				  $$->val = $1->val;
				}
	| TIMEOUT		{ $$ = nn(ZN,TIMEOUT, ZN, ZN); }
	| NONPROGRESS		{ $$ = nn(ZN,NONPROGRESS, ZN, ZN);
				  has_np++;
				}
	| PC_VAL '(' expr ')'	{ $$ = nn(ZN, PC_VAL, $3, ZN);
				  has_pcvalue++;
				}
	| PNAME '[' expr ']' '@'
	  NAME			{ $$ = rem_lab($1->sym, $3, $6->sym); }
	;


new:    NEW typeref                 { $$ = nn($2, NEW, ZN, ZN); }
      | NEW typeref '[' expr ']'    { $$ = nn($2, NEW, $4, ZN); }
      ;


rettype:  /* empty */           { $$ = ZN; }
         | ':' typeref          { $$ = $2; }
	 ;


typeref: TYPE                   { $$ = nn(ZN, TYPE, ZN, ZN);
                                  $$->sym = addsym(type_name($1->val));
				  $$->sym->type = $1->val;

				  /* careful with symbol duplication */
				  $$->sym->hidden |= 128;
                                }

         | UNAME                { $$ = nn(ZN, TYPE, ZN, ZN);

	                          /* make sure we get the right one */
	                          $$->sym = getusym($1->sym);
	                        }

         | typeref '&'          { $$ = nn(ZN, TYPE, ZN, ZN);
	                          $$->sym = addsym($1->sym->name);
				  $$->sym->type = $1->sym->type | POINTER;

				  /* careful with symbol duplication */
				  $$->sym->hidden |= 128;
	                        }
         ;

Opt_priority:	/* none */	{ $$ = ZN; }
	| PRIORITY CONST	{ $$ = $2; }
	;

full_expr:	expr		{ $$ = $1; }
	|	Expr		{ $$ = $1; }
	;

Opt_enabler:	/* none */	{ $$ = ZN; }
	| PROVIDED '(' full_expr ')'	{ if (!proper_enabler($3))
				  {	non_fatal("invalid PROVIDED clause",
						(char *)0);
					$$ = ZN;
				  } else
					$$ = $3;
				 }
	| PROVIDED error	{ $$ = ZN;
				  non_fatal("usage: provided ( ..expr.. )",
					(char *)0);
				}
	;

	/* an Expr cannot be negated - to protect Probe expressions */
Expr	: Probe			{ $$ = $1; }
	| '(' Expr ')'		{ $$ = $2; }
	| Expr AND Expr		{ $$ = nn(ZN, AND, $1, $3); }
	| Expr AND expr		{ $$ = nn(ZN, AND, $1, $3); }
	| Expr OR  Expr		{ $$ = nn(ZN,  OR, $1, $3); }
	| Expr OR  expr		{ $$ = nn(ZN,  OR, $1, $3); }
	| expr AND Expr		{ $$ = nn(ZN, AND, $1, $3); }
	| expr OR  Expr		{ $$ = nn(ZN,  OR, $1, $3); }
	;

Probe	: FULL '(' varref ')'	{ $$ = nn($3,  FULL, $3, ZN); }
	| NFULL '(' varref ')'	{ $$ = nn($3, NFULL, $3, ZN); }
	| EMPTY '(' varref ')'	{ $$ = nn($3, EMPTY, $3, ZN); }
	| NEMPTY '(' varref ')'	{ $$ = nn($3,NEMPTY, $3, ZN); }
	;

/*
 * This non-terminal only mimicks typeref in a simplistic way.
 * One day it should be removed, and typeref re-engineered.
 */
basetype: TYPE			{ $$->sym = ZS;
				  $$->val = $1->val;
				  if ($$->val == UNSIGNED)
				  fatal("unsigned cannot be used as mesg type",
					0);
				}
	| UNAME			{ $$->sym = $1->sym;
				  $$->val = STRUCT;
				}
        | basetype '&'          { $$->sym = $1->sym;
	                          $$->val |= POINTER;
	                        }

	| error			/* e.g., unsigned ':' const */
	;

typ_list: basetype		{ $$ = nn($1, $1->val, ZN, ZN); }
	| basetype ',' typ_list	{ $$ = nn($1, $1->val, ZN, $3); }
	;

args    : /* empty */		{ $$ = ZN; }
	| arg			{ $$ = $1; }
	;

prargs  : /* empty */		{ $$ = ZN; }
	| ',' arg		{ $$ = $2; }
	;

margs   : arg			{ $$ = $1; }
	| expr '(' arg ')'	{ if ($1->ntyp == ',')
					$$ = tail_add($1, $3);
				  else
				  	$$ = nn(ZN, ',', $1, $3);
				}
	;

arg     : expr			{ if ($1->ntyp == ',')
					$$ = $1;
				  else
				  	$$ = nn(ZN, ',', $1, ZN);
				}

	| expr ',' arg		{ if ($1->ntyp == ',')
					$$ = tail_add($1, $3);
				  else
				  	$$ = nn(ZN, ',', $1, $3);
				}
	;

rarg	: varref		{ $$ = $1; trackvar($1, $1);
				  trapwonly($1, "rarg"); }

	| EVAL '(' varref ')'	{ $$ = nn(ZN,EVAL,$3,ZN);
				  trapwonly($1, "eval rarg"); }

	| CONST 		{ $$ = nn(ZN,CONST,ZN,ZN);
				  $$->ismtyp = $1->ismtyp;
				  $$->val = $1->val;
				}

	| '-' CONST %prec UMIN	{ $$ = nn(ZN,CONST,ZN,ZN);
				  $$->val = - ($2->val);
				}
	;

rargs	: rarg			{ if ($1->ntyp == ',')
					$$ = $1;
				  else
				  	$$ = nn(ZN, ',', $1, ZN);
				}
	| rarg ',' rargs	{ if ($1->ntyp == ',')
					$$ = tail_add($1, $3);
				  else
				  	$$ = nn(ZN, ',', $1, $3);
				}
	| rarg '(' rargs ')'	{ if ($1->ntyp == ',')
					$$ = tail_add($1, $3);
				  else
				  	$$ = nn(ZN, ',', $1, $3);
				}
	| '(' rargs ')'		{ $$ = $2; }
	;

nlst	: NAME			{ $$ = nn($1, NAME, ZN, ZN);
				  $$ = nn(ZN, ',', $$, ZN); }

	| nlst NAME 		{ $$ = nn($2, NAME, ZN, ZN);
				  $$ = nn(ZN, ',', $$, $1);
				}

	| nlst ','		{ $$ = $1; /* commas optional */ }
	;
%%

void
yyerror(char *fmt, ...)
{
  parse_errors ++;
  non_fatal(fmt, (char *) 0);
}
