/***** spin: structs.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif


typedef struct SNode 
{
  Symbol        *sym;
  struct SNode  *nxt;
} SNode;

static SNode * SList = NULL; 

extern	Symbol	*context;
extern	RunList	*X;
extern	Symbol	*Fname;
extern	int	lineno, depth;
extern  int     Expand_Ok;

Symbol	*owner;
UType   *Unames = 0;

static UType *Pnames = 0;
static int    U_nid  = 10000; /* hopefully big enough */

extern void	sr_mesg(FILE *, int, int);

extern int get_val(Symbol *, int);
extern int get_lval(Symbol *, Lextok *);
extern Lextok *get_Sval(Symbol *, int);
extern void set_val(Symbol *, int, int);
extern void set_lval(Symbol *, Lextok *, int);
extern void set_Sval(Symbol *, int, Lextok *);


void
setuname(Lextok *n)
{	
  UType *tmp;
  
  if (!owner) 
    fatal("illegal reference inside typedef", (char *) 0);

  for (tmp = Unames; tmp; tmp = tmp->nxt)
    if (!strcmp(owner->name, tmp->nm->name))
      break;

  if (!tmp) {
    tmp = (UType *)emalloc(sizeof(UType));
    tmp->nm = owner;
    tmp->nxt = Unames;
    Unames = tmp;
  } else
    if (tmp->cn)
      fatal("type '%s' already defined", owner->name);
  
  tmp->cn = n;

  
}


static void
bp_flds(Symbol *s)
{
  Lextok *fp, *tl;

  for (fp = s->Slst; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      {
	if (tl->sym->type & STRUCT
	    && !strcmp(tl->sym->Snm->name, s->name))
	  tl->sym->Slst = s->Slst;
      }
}


void
check_unames()
{
  UType *tmp;
  Symbol *s;
  int i;

  for (tmp = Unames; tmp; tmp = tmp->nxt)
    {
      if (!tmp->cn)
	fatal("Undefined type '%s'", tmp->nm->name);

      tmp->nm->Nid = U_nid ++;
      tmp->nm->type = STRUCT;
      tmp->nm->Slst = tmp->cn;
      tmp->nm->hidden |= 128;

      for (i = 0; i < tmp->ps_no; i ++)
	{
	  s = tmp->ps_lst [i];
	  s->Slst = tmp->cn;
	}

      bp_flds(tmp->nm);
    }
}


int
isutype(char *t)
{	UType *tmp;

	for (tmp = Unames; tmp; tmp = tmp->nxt)
	  {	
	    if (!strcmp(t, tmp->nm->name))
	      return 1;
	  }
	
	return 0;
}


Lextok *
getuname(Symbol *t)
{	
  UType *tmp;
	
  for (tmp = Unames; tmp; tmp = tmp->nxt)
    {	
      if (!strcmp(t->name, tmp->nm->name))
	return tmp->cn;
    }
	
  fatal("%s is not a typename (getuname)", t->name);
  return ZN;
}


Symbol *
getusym(Symbol *t)
{
  UType *tmp;

  for (tmp = Unames; tmp; tmp = tmp->nxt)
    {
      if (!strcmp(t->name, tmp->nm->name))
	return tmp->nm;
    }

  fatal("%s is not a typename (getusym)", t->name);
  return ZS;
}


static void 
postpone(Symbol *t, Symbol *s)
{
  UType *tmp;

  for (tmp = Unames; tmp; tmp = tmp->nxt)
    {
      if (!strcmp(t->name, tmp->nm->name))
	{
	  tmp->ps_no ++;
	  tmp->ps_lst = (Symbol **) erealloc(tmp->ps_lst, 
					     tmp->ps_no * sizeof(Symbol *));
	  tmp->ps_lst [tmp->ps_no - 1] = s;
	}
    }
}


void
setutype(Lextok *p, Symbol *t, Lextok *vis)	/* user-defined types */
{	
  int oln = lineno;
  Symbol *ofn = Fname;
  Lextok *m, *n;

  if (owner 
      && !strcmp(owner->name, t->name) 
      && !(p->sym->type & POINTER))
    fatal("incomplete type field", 0);

  m = getuname(t);
  
  for (n = p; n; n = n->rgt)
    {	
      lineno = n->ln;
      Fname = n->fn;

      if (n->sym->type & ~POINTER)
	non_fatal("redeclaration of '%s'", n->sym->name);
      
      if (n->sym->nbits > 0)
	non_fatal("(%s) only an unsigned can have width-field",
		  n->sym->name);

      if (vis)
	{	
	  if (strncmp(vis->sym->name, ":hide:", 6) == 0)
	    n->sym->hidden |= 1;
	  
	  if (strncmp(vis->sym->name, ":show:", 6) == 0)
	    n->sym->hidden |= 2;
	}

      n->sym->type |= STRUCT;	/* classification   */

      if (m)
	n->sym->Slst = m;	/* structure itself */
      else
	postpone(t, n->sym);

      n->sym->Snm  = t;	/* name of typedef  */
      n->sym->Nid  = 0;	/* this is no chan  */
      n->sym->hidden |= 4;
      
      if (n->sym->nel <= 0)
	non_fatal("bad array size for '%s'", n->sym->name);
    }
	
  lineno = oln;
  Fname = ofn;
}


static Symbol *
do_same(Lextok *n, Symbol *v)
{	
  Lextok *tmp, *fp, *tl, *Sval;
  int oln = lineno;
  Symbol *ofn = Fname;
  int ix;

  lineno = n->ln;
  Fname = n->fn;

  ix = eval(n->lft);
  
  if (!(v->type & STRUCT))
    return v;

  /* n->sym->type & STRUCT
   * index:		 n->lft
   * subfields:		 n->rgt
   * structure template: n->sym->Slst
   * runtime values:	 n->sym->Sval
   */
  
  if (!((v->type & POINTER) && v->nel == 1)
      && (ix >= v->nel || ix < 0))
    {	
      printf("spin: indexing %s[%d] - size is %d\n", v->name, ix, v->nel);
      fatal("indexing error \'%s\'", v->name);
    }

  if (Expand_Ok || (v->type & POINTER)) 
    {
      if (!n->rgt || !n->rgt->lft)
	return v;
    }

  if ((v->type & POINTER) && !v->Sval)    
    fatal("dereferencing null pointer '%s'", v->name);

  if (!n->rgt || !n->rgt->lft)
    fatal("no subfields '%s'", v->name);  /* i.e., wants all */

  if (n->rgt->ntyp != '.')
    {	
      printf("bad subfield type %d\n", n->rgt->ntyp);
      alldone(1);
    }

  tmp = n->rgt->lft;
  if (tmp->ntyp != NAME && tmp->ntyp != TYPE)
    {	
      printf("bad subfield entry %d\n", tmp->ntyp);
      alldone(1);
    }

  if ((v->type & POINTER) && v->nel > 1)
    {
      if (!(Lextok **)v->Sval[ix + 1]) {
	char buf[256];
	sprintf(buf, "dereferencing null pointer '%s[%d]'", v->name, ix);
	fatal(buf, 0);
      }

      Sval = ((Lextok **)v->Sval[ix + 1]) [1];
    }
  else
    Sval = v->Sval[ix + 1];

  for (fp = Sval; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      if (!strcmp(tl->sym->name, tmp->sym->name))
	{	
	  lineno = oln; Fname = ofn;
	  return tl->sym;
	}

  fatal("cannot locate subfield '%s'", tmp->sym->name);
  return ZS;
}


int
Rval_struct(Lextok *n, Symbol *v)	/* n varref, v valref */
{	
  Symbol *tl;
  Lextok *tmp;
  int ix;
  lineno = n->ln;
  Fname  = n->fn;

  if (!n || !(tl = do_same(n, v)))
    return 0;

  /* 
   * CHANGE: structure pointers uncompletely dereferenced
   *         evaluate to their left value.
   */
  if (!n->rgt || 
      !n->rgt->lft)
    return (int)tl->Sval;

  tmp = n->rgt->lft;
  
  if (tmp->sym->type & STRUCT)
    return Rval_struct(tmp, tl);
  else if (tmp->rgt)
    fatal("non-zero 'rgt' on non-structure", 0);
  
  ix = eval(tmp->lft);  
  if (!((tl->type & POINTER) && tl->nel == 1) 
      && (ix >= tl->nel || ix < 0))
    {
      printf("spin: indexing %s[%d] - size is %d\n", tl->name, ix, tl->nel);
      fatal("indexing error \'%s\'", tl->name);
    }

  if (tl->type == FTYPE)
    return (int)get_Sval(tl, ix);
  else
    return cast_val(tl->type, get_val(tl, ix), tl->nbits);
}


int 
Rval_lstruct(Lextok *n, Symbol *v)
{
  Symbol *tl;

  if (n->rgt)
    if (!(tl = do_same(n, v)))
      return 1;
    else
      return Rval_lstruct(n->rgt->lft, tl);

  if (v->type & STRUCT)
    {
      if (n->lft)
	return (int)get_Sval(v, eval(n->lft));
      else
	return (int)v->Sval;
    }
  else
    return get_lval(v, n->lft);
}


int
Lval_struct(Lextok *n, Symbol *v, int a)  /* a = assigned value */
{	
  Lextok *tmp;
  Symbol *tl;
  int ix;
	
  if (!(tl = do_same(n, v)))
    return 1;
  
  if (!n->rgt ||
      !n->rgt->lft)
    {
      ix = eval(n->lft);      
      memcpy((char *)(tl->Sval + ix), (char *)a, sizeof(Lextok *));
      return 1;
    }

  tmp = n->rgt->lft;
  
  if (tmp->sym->type & STRUCT)
    return Lval_struct(tmp, tl, a);
  else if (tmp->rgt && tmp->rgt->lft)
    fatal("no structure '%s'", v->name);
  
  ix = eval(tmp->lft);

  if (!(tl->type & POINTER && tl->nel == 1) 
      && (ix >= tl->nel || ix < 0))
    {
      printf("spin: indexing %s[%d] - size is %d\n", tl->name, ix, tl->nel);
      fatal("indexing error \'%s\'", tl->name);
    }

  if (tl->nbits > 0)
    a = (a & ((1<<tl->nbits)-1));
  
  set_val(tl, ix, a);
  tl->setat = depth;
  
  return 1;
}


int
Lval_lstruct(Lextok *n, Symbol *v, int a, int heap)
{
  Symbol *tl;
  Lextok *tmp;

  if (n->rgt) /* subfield expression */
    if (!(tl = do_same(n, v)))
      return 1;
    else
      return Lval_lstruct(n->rgt->lft, tl, a, heap);
  
  if (v->type & STRUCT)
    {
      if (n->lft)
	set_Sval(v, eval(n->lft), (Lextok *)a);
      else
	v->Sval = (Lextok **)a;
    }
  else
    set_lval(v, n->lft, a);
  
  if (heap)
    v->hidden &= 64;

  return 1;
}


int
Lval_fstruct(Lextok *n, Symbol *v, Sequence *body)
{
  Symbol *tl;
  Lextok *tmp;
  int ix;

  if (!(tl = do_same(n, v)))
    return 1;

  tmp = n->rgt->lft;

  if (tmp->sym->type & STRUCT)
    return Lval_fstruct(tmp, tl, body);
  else
    if (tmp->rgt)
      fatal("non-zero 'rgt' on non-structure", 0);

  ix = eval(tmp->lft);
  if (!(tl->type & POINTER && tl->nel == 1)
      && (ix >= tl->nel || ix < 0))
    {
      printf("spin: indexing %s[%d] - size is %d\n", tl->name, ix, tl->nel);
      fatal("indexing error \'%s\'", tl->name);
    }

  set_Sval(tl, ix, (Lextok *)body);
  return 1;
}


int
Cnt_flds(Lextok *m)
{	
  Lextok *fp, *tl, *n;
  int cnt = 0;

  if (m->ntyp == ',')
    {	
      n = m;
      goto is_lst;
    }

  if (!m->sym || m->ntyp != STRUCT)
    return 1;

  n = getuname(m->sym);
is_lst:
  for (fp = n; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      {	
	if (tl->sym->type == STRUCT)
	  {	
	    if (tl->sym->nel > 1)
	      fatal("hidden array in parameter, %s",
		    tl->sym->name);
	    cnt += Cnt_flds(tl->sym->Slst);
	  }  else
	    cnt += tl->sym->nel;
      }
  return cnt;
}


Symbol *
struct_sym(Lextok *n, Symbol *s)
{
  Symbol *v;

  if (!s)
    return ZS;

  /* get the right symbol */
  if (!s->owner && (v = findloc(s)))
    s = v;

  if (!(v = do_same(n, s)))
    return ZS;

  if (!(v->type & STRUCT))
    return v;

  if (!(Expand_Ok || v->type & POINTER))
    {
      if (!n->rgt 
	  || !n->rgt->ntyp == '.' 
	  || !n->rgt->lft)
	fatal("unexpected struct layout `%s`", v->name);

      return struct_sym(n->rgt->lft, v);
    }
  
  if (!n->rgt)
    return v;
  
  return struct_sym(n->rgt->lft, v);
}


Lextok *
struct_tok(Lextok *n, Symbol *s)
{
  Symbol *v;

  if (!s)
    return ZN;

    /* get the right symbol */
  if (!s->owner && (v = findloc(s)))
    s = v;

  if (!(v = do_same(n, s)))
    return ZN;

  if (!n->rgt)
    return n;
  
  if (!(Expand_Ok || v->type & POINTER))
    {
      if (!n->rgt->ntyp == '.' 
	  || !n->rgt->lft)
	fatal("unexpected struct layout `%s`", v->name);

      return struct_tok(n->rgt->lft, v);
    }

  return struct_tok(n->rgt->lft, v);
}


int
Width_set(int *wdth, int i, Lextok *n)
{	
  Lextok *fp, *tl;
  int j = i, k;

  for (fp = n; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      {	
	if (tl->sym->type == STRUCT)
	  j = Width_set(wdth, j, tl->sym->Slst);
	else
	  {	
	    for (k = 0; k < tl->sym->nel; k++, j++)
	      wdth[j] = tl->sym->type;
	  }	
      }
  
  return j;
}


void
ini_struct(Symbol *s)
{	
  int i; 
  Lextok *fp, *tl;
  
  if (s->type != STRUCT)	/* last step */
    {	
      ini_var(s);
      return;
    }
  
  if (!s->Sval)
    s->Sval = (Lextok **) emalloc((s->nel + 1) * sizeof(Lextok *));

  s->Sval[0] = (Lextok *) s->nel;
  for (i = 0; i < s->nel; i++)
    {	
      s->Sval[i + 1] = cpnn(s->Slst, 1, 1, 1);
      for (fp = s->Sval[i + 1]; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  ini_struct(tl->sym);
    }	
}


Lextok *
cpnn(Lextok *s, int L, int R, int S)
{	
  Lextok *d; 
  extern int Nid;

  if (!s) 
    return ZN;

  d = (Lextok *) emalloc(sizeof(Lextok));

  d->ntyp = s->ntyp;
  d->val  = s->val;
  d->ln   = s->ln;
  d->fn   = s->fn;
  d->sym  = s->sym;

  if (L) 
    d->lft = cpnn(s->lft, 1, 1, S);
	
  if (R) 
    d->rgt = cpnn(s->rgt, 1, 1, S);

  if (S && s->sym)
    {	
      d->sym = (Symbol *) emalloc(sizeof(Symbol));
      memcpy(d->sym, s->sym, sizeof(Symbol));

      if (d->sym->type == CHAN)
	d->sym->Nid = ++ Nid;
    }
  
  if (s->sq || s->sl)
    fatal("cannot happen cpnn", (char *) 0);

  return d;
}


void
validref(Lextok *p, Lextok *c)
{	
  Lextok *fp, *tl;
  char lbuf[512];
  Symbol *s = lookup(p->sym->name);

  for (fp = s->Slst; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      if (strcmp(tl->sym->name, c->sym->name) == 0)
	return;

  sprintf(lbuf, "no field '%s' defined in structure '%s'\n",
	  c->sym->name, p->sym->name);

  fatal(lbuf, (char *) 0);
}

void
struct_name(Lextok *n, Symbol *v, char *buf)
{	Symbol *tl;
	Lextok *tmp;
	char lbuf[128];

	if (!n || !(tl = do_same(n, v)))
		return;
	tmp = n->rgt->lft;
	if (tmp->sym->type == STRUCT)
	{	strcat(buf, ".");
		struct_name(tmp, tl, buf);
		return;
	}
	sprintf(lbuf, ".%s", tl->name);
	strcat(buf, lbuf);
	if (tmp->sym->nel > 1)
	{	
	  sprintf(lbuf, "[%d]", eval(tmp->lft));
	  strcat(buf, lbuf);
	}
}


static void 
Sadd(Symbol *s)
{
  SNode *n = (SNode *) emalloc (sizeof(SNode));
  
  n->sym = s;
  n->nxt = SList;
  SList  = n; 
}


static int 
Smem(Symbol *s)
{
  SNode *n;

  for (n = SList; n; n = n->nxt)
    {
      if (n->sym == s)
	return 1;
    }
  
  return 0;
}


void
Sclear()
{
  SNode *n;

  while (SList)
    {
      n = SList;
      SList = SList->nxt;
      
    }
}


void
dump_struct(Symbol *z, char *prefix, RunList *r)
{	
  Lextok *fp, *tl, *Sval;
  char eprefix[128];
  int ix, jx;

  if (Smem(z))
    return;

  Sadd(z);

  for (ix = 0; ix < z->nel; ix ++)
    {	
      if (z->nel > 1)
	sprintf(eprefix, "%s[%d]", prefix, ix);
      else
	strcpy(eprefix, prefix);

      if (z->type & POINTER && z->nel > 1)
	Sval = z->Sval[ix + 1] ? ((Lextok **) z->Sval[ix + 1]) [1] : ZN;
      else
	Sval = z->Sval[ix + 1];
    
      for (fp = Sval; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  {	
	    if (tl->sym->type & STRUCT
		&& tl->sym->Sval)
	      {	
		char pref[128];

		strcpy(pref, eprefix);
		strcat(pref, ".");
		strcat(pref, tl->sym->name);

		dump_struct(tl->sym, pref, r);
	      } else
		for (jx = 0; jx < tl->sym->nel; jx++)
		  {	
		    if (tl->sym->type == CHAN)
		      doq(tl->sym, jx, r);
		    else
		      {	
			printf("\t\t");

			if (r)
			  printf("%s(%d):", r->n->name, r->pid);

			printf("%s.%s", eprefix, tl->sym->name);
			
			if (tl->sym->nel > 1)
			  printf("[%d]", jx);
			
			printf(" = ");

			if (tl->sym->val)
			  sr_mesg(stdout, tl->sym->val[jx + 1], 
				  tl->sym->type == MTYPE);
			else
			  printf("0");

			printf("\n");
		      }	
		  }	
	  }
    }
}


static int
retrieve(Lextok **targ, int i, int want, Lextok *n, int Ntyp)
{	Lextok *fp, *tl;
	int j = i, k;

	for (fp = n; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	{	if (tl->sym->type & STRUCT)
		{	j = retrieve(targ, j, want, tl->sym->Slst, Ntyp);
			if (j < 0)
			{	Lextok *x = cpnn(tl, 1, 0, 0);
				x->rgt = nn(ZN, '.', (*targ), ZN);
				(*targ) = x;
				return -1;
			}
		} else
		{	for (k = 0; k < tl->sym->nel; k++, j++)
			{	if (j == want)
				{	*targ = cpnn(tl, 1, 0, 0);
					(*targ)->lft = nn(ZN, CONST, ZN, ZN);
					(*targ)->lft->val = k;
					if (Ntyp)
					(*targ)->ntyp = (short) Ntyp;
					return -1;
				}
	}	}	}
	return j;
}

static int
is_explicit(Lextok *n)
{
	if (!n) return 0;
	if (!n->sym) fatal("unexpected - no symbol", 0);
	if (!(n->sym->type & STRUCT)) return 1;

	/* CHANGE: pointers to structures are explicit */
	if ((n->sym->type & (POINTER|STRUCT)) && !n->rgt) return 1;

	if (!n->rgt) return 0;

	if (n->rgt->ntyp != '.')
	{	lineno = n->ln;
		Fname  = n->fn;
		fatal("unexpected %s, no '.'", n->sym->name);
	}
	return is_explicit(n->rgt->lft);
}

Lextok *
expand(Lextok *n, int Ok)
	/* turn rgt-lnked list of struct nms, into ',' list of flds */
{	Lextok *x = ZN, *y;

	if (!Ok) return n;

	while (n)
	{	y = mk_explicit(n, 1, 0);
		if (x)
		  (void) tail_add(x, y);
		else
		  x = y;

		n = n->rgt;
	}

	return x;
}

Lextok *
mk_explicit(Lextok *n, int Ok, int Ntyp)
	/* produce a single ',' list of fields */
{	Lextok *bld = ZN, *x;
	int i, cnt; 
	extern int IArgs;
	extern int LValue;

	if (is_explicit(n))
	  return n;

	if (!Ok || !n->sym->Slst)
	  {	if (IArgs) return n;
	        printf("spin: saw '");
	        comment(stdout, n, 0);
		printf("'\n");
		fatal("incomplete structure reference '%s'", n->sym->name);
	  }

	cnt = Cnt_flds(n->sym->Slst);
	for (i = cnt-1; i >= 0; i--)
	{	bld = nn(ZN, ',', ZN, bld);
		if (retrieve(&(bld->lft), 0, i, n->sym->Slst, Ntyp) >= 0)
		{	printf("cannot retrieve field %d\n", i);
			fatal("bad structure %s", n->sym->name);
		}
		x = cpnn(n, 1, 0, 0);
		x->rgt = nn(ZN, '.', bld->lft, ZN);
		bld->lft = x;
	}

	bld->sym = n->sym;
	return bld;
}

Lextok *
tail_add(Lextok *a, Lextok *b)
{	Lextok *t;

	for (t = a; t->rgt; t = t->rgt)
		if (t->ntyp != ',')
		fatal("unexpected type - tail_add", 0);
	t->rgt = b;
	return a;
}

void
setpname(Lextok *n)
{	UType *tmp;

	for (tmp = Pnames; tmp; tmp = tmp->nxt)
		if (!strcmp(n->sym->name, tmp->nm->name))
		{	non_fatal("proctype %s redefined",
				n->sym->name);
			return;
		}
	tmp = (UType *) emalloc(sizeof(UType));
	tmp->nm = n->sym;
	tmp->nxt = Pnames;
	Pnames = tmp;
}

int
isproctype(char *t)
{	UType *tmp;

	for (tmp = Pnames; tmp; tmp = tmp->nxt)
	{	if (!strcmp(t, tmp->nm->name))
			return 1;
	}
	return 0;
}
